import datetime as dt
import random

table_with_labels_data = [
    {
        "date": dt.date(2021, 1, 1),
        "x": 5,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Ana",
        "name2": "Ana",
        "name3": "Ana",
    },
    {
        "date": dt.date(2021, 1, 2),
        "x": 6,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "Z",
        "name": "Laura",
        "name2": "Laura",
        "name3": "Laura",
    },
    {
        "date": dt.date(2021, 1, 3),
        "x": 4,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "W",
        "name": "Audrey",
        "name2": "Audrey",
        "name3": "Audrey",
    },
    {
        "date": dt.date(2021, 1, 4),
        "x": 7,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "W",
        "name": "Jose",
        "name2": "Jose",
        "name3": "Jose",
    },
    {
        "date": dt.date(2021, 1, 5),
        "x": 3,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Jorge",
        "name2": "Jorge",
        "name3": "Jorge",
    },
    {
        "date": dt.date(2021, 1, 1),
        "x": 5,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Ana",
        "name2": "Ana",
        "name3": "Ana",
    },
    {
        "date": dt.date(2021, 1, 2),
        "x": 6,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "Z",
        "name": "Laura",
        "name2": "Laura",
        "name3": "Laura",
    },
    {
        "date": dt.date(2021, 1, 3),
        "x": 4,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "W",
        "name": "Audrey",
        "name2": "Audrey",
        "name3": "Audrey",
    },
    {
        "date": dt.date(2021, 1, 4),
        "x": 7,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "W",
        "name": "Jose",
        "name2": "Jose",
        "name3": "Jose",
    },
    {
        "date": dt.date(2021, 1, 5),
        "x": 3,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Jorge",
        "name2": "Jorge",
        "name3": "Jorge",
    },
    {
        "date": dt.date(2021, 1, 1),
        "x": 5,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Ana",
        "name2": "Ana",
        "name3": "Ana",
    },
    {
        "date": dt.date(2021, 1, 2),
        "x": 6,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "Z",
        "name": "Laura",
        "name2": "Laura",
        "name3": "Laura",
    },
    {
        "date": dt.date(2021, 1, 3),
        "x": 4,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "W",
        "name": "Audrey",
        "name2": "Audrey",
        "name3": "Audrey",
    },
    {
        "date": dt.date(2021, 1, 4),
        "x": 7,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "W",
        "name": "Jose",
        "name2": "Jose",
        "name3": "Jose",
    },
    {
        "date": dt.date(2021, 1, 5),
        "x": 3,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Jorge",
        "name2": "Jorge",
        "name3": "Jorge",
    },
    {
        "date": dt.date(2021, 1, 1),
        "x": 5,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Ana",
        "name2": "Ana",
        "name3": "Ana",
    },
    {
        "date": dt.date(2021, 1, 2),
        "x": 6,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "Z",
        "name": "Laura",
        "name2": "Laura",
        "name3": "Laura",
    },
    {
        "date": dt.date(2021, 1, 3),
        "x": 4,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "W",
        "name": "Audrey",
        "name2": "Audrey",
        "name3": "Audrey",
    },
    {
        "date": dt.date(2021, 1, 4),
        "x": 7,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "W",
        "name": "Jose",
        "name2": "Jose",
        "name3": "Jose",
    },
    {
        "date": dt.date(2021, 1, 5),
        "x": 3,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Jorge",
        "name2": "Jorge",
        "name3": "Jorge",
    },
    {
        "date": dt.date(2021, 1, 1),
        "x": 5,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Ana",
        "name2": "Ana",
        "name3": "Ana",
    },
    {
        "date": dt.date(2021, 1, 2),
        "x": 6,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "Z",
        "name": "Laura",
        "name2": "Laura",
        "name3": "Laura",
    },
    {
        "date": dt.date(2021, 1, 3),
        "x": 4,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "W",
        "name": "Audrey",
        "name2": "Audrey",
        "name3": "Audrey",
    },
    {
        "date": dt.date(2021, 1, 4),
        "x": 7,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "W",
        "name": "Jose",
        "name2": "Jose",
        "name3": "Jose",
    },
    {
        "date": dt.date(2021, 1, 5),
        "x": 3,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Jorge",
        "name2": "Jorge",
        "name3": "Jorge",
    },
    {
        "date": dt.date(2021, 1, 1),
        "x": 5,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Ana",
        "name2": "Ana",
        "name3": "Ana",
    },
    {
        "date": dt.date(2021, 1, 2),
        "x": 6,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "Z",
        "name": "Laura",
        "name2": "Laura",
        "name3": "Laura",
    },
    {
        "date": dt.date(2021, 1, 3),
        "x": 4,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "W",
        "name": "Audrey",
        "name2": "Audrey",
        "name3": "Audrey",
    },
    {
        "date": dt.date(2021, 1, 4),
        "x": 7,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "B",
        "filtB": "W",
        "name": "Jose",
        "name2": "Jose",
        "name3": "Jose",
    },
    {
        "date": dt.date(2021, 1, 5),
        "x": 3,
        "y": round(100 * random.random(), 10),
        "z": round(100 * random.random(), 1),
        "a": round(100 * random.random(), 1),
        "filtA": "A",
        "filtB": "Z",
        "name": "Jorge",
        "name2": "Jorge",
        "name3": "Jorge",
    },
]



input_form_data = {
    "fields": [
        {
            "title": "Personal information",
            "fields": [
                {
                    "mapping": "name",
                    "fieldName": "name",
                    "inputType": "text",
                },
                {
                    "mapping": "surname",
                    "fieldName": "surname",
                    "inputType": "text",
                },
                {
                    "mapping": "age",
                    "fieldName": "age",
                    "inputType": "number",
                },
                {
                    "mapping": "tel",
                    "fieldName": "phone",
                    "inputType": "tel",
                },
                {
                    "mapping": "gender",
                    "fieldName": "Gender",
                    "inputType": "radio",
                    "options": ["Male", "Female", "No-binary", "Undefined"],
                },
                {
                    "mapping": "email",
                    "fieldName": "email",
                    "inputType": "email",
                },
            ],
        },
        {
            "title": "Other data",
            "fields": [
                {
                    "mapping": "skills",
                    "fieldName": "Skills",
                    "options": [
                        "Backend",
                        "Frontend",
                        "UX/UI",
                        "Api Builder",
                        "DevOps",
                    ],
                    "inputType": "checkbox",
                },
                {
                    "mapping": "birthDay",
                    "fieldName": "Birthday",
                    "inputType": "date",
                },
                {
                    "mapping": "onCompany",
                    "fieldName": "time on Shimoku",
                    "inputType": "dateRange",
                },
                {
                    "mapping": "hobbies",
                    "fieldName": "Hobbies",
                    "inputType": "select",
                    "options": [
                        "Make Strong Api",
                        "Sailing to Canarias",
                        "Send Abracitos",
                    ],
                },
                {
                    "mapping": "textField2",
                    "fieldName": "Test Text",
                    "inputType": "text",
                },
            ],
        },
    ],
}



from os import getenv
from shimoku.exceptions import CacheError, APIError
import unittest
from utils import initiate_shimoku

s = initiate_shimoku()

business_id: str = getenv("BUSINESS_ID")
environment: str = getenv("ENVIRONMENT")
mock: bool = getenv("MOCK") == "TRUE"
playground: bool = getenv("PLAYGROUND") == "TRUE"

s.set_workspace(uuid=business_id)


def delete_dashboard_if_exists(name):
    if s.boards.get_board(name=name):
        s.boards.force_delete_board(name=name)


class TestDashboardMetadataApi(unittest.TestCase):
    def test_dashboard_CRUD(self):
        name = "Testing dashboard CRUD"
        delete_dashboard_if_exists(name)
        delete_dashboard_if_exists(name + " updated")

        dashboard = s.boards.create_board(name=name)

        assert dashboard["name"] == name

        dashboard = s.boards.get_board(uuid=dashboard["id"])
        assert dashboard["name"] == name

        s.boards.update_board(
            uuid=dashboard["id"], new_name=name + " updated", is_public=True
        )

        dashboard = s.boards.get_board(uuid=dashboard["id"])
        assert dashboard["name"] == name + " updated"

        assert s.boards.create_board(name=name)
        with self.assertRaises(CacheError):
            s.boards.create_board(name=name)
        with self.assertRaises(CacheError):
            s.boards.create_board(name=name + " updated")

        s.boards.delete_board(name=name)
        s.boards.delete_board(name=name + " updated")

        assert not s.boards.get_board(name=name)
        assert not s.boards.get_board(name=name + " updated")

    def test_create_and_delete_app_dashboard_link(self):
        name = "Testing dashboard for appdashboards"

        app_name = "Testing app"
        s.menu_paths.delete_menu_path(
            uuid=s.menu_paths.get_menu_path(name=app_name)["id"]
        )

        delete_dashboard_if_exists(name)
        delete_dashboard_if_exists(name + " updated")

        s.boards.create_board(name=name)
        app_id = s.menu_paths.get_menu_path(name=app_name)["id"]

        s.boards.add_menu_path_in_board(name=name, menu_path_id=app_id)

        app_ids = s.boards.get_board_menu_path_ids(name=name)
        assert app_id == app_ids[0]

        if not mock and not playground:
            with self.assertRaises(APIError):
                s.boards.delete_board(name=name)

        s.boards.remove_menu_path_from_board(name=name, menu_path_id=app_id)
        s.boards.delete_board(name=name)
        s.menu_paths.delete_menu_path(uuid=app_id)

    def test_delete_all_app_dashboards_links(self):
        name = "Testing dashboard for appdashboards delete all"

        app_names = ["Testing app 1", "Testing app 2", "Testing app 3"]

        for app_name in app_names:
            s.menu_paths.delete_menu_path(
                uuid=s.menu_paths.get_menu_path(name=app_name)["id"]
            )

        delete_dashboard_if_exists(name)

        app_ids = []
        s.boards.create_board(name=name)
        for app_name in app_names:
            app_id = s.menu_paths.get_menu_path(name=app_name)["id"]
            s.boards.add_menu_path_in_board(name=name, menu_path_id=app_id)
            app_ids.append(app_id)

        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)
        assert len(dashboard_app_ids) == len(app_names)

        s.boards.remove_all_menu_paths_from_board(name=name)

        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)
        assert len(dashboard_app_ids) == 0

        for app_id in app_ids:
            s.boards.add_menu_path_in_board(name=name, menu_path_id=app_id)

        s.boards.force_delete_board(name=name)

        assert not s.boards.get_board(name=name)

        for app_id in app_ids:
            s.menu_paths.delete_menu_path(uuid=app_id)

    def test_group_apps(self):
        name = "Testing dashboard for grouping apps"

        app_names = ["Testing app 1", "Testing app 2", "Testing app 3"]

        for app_name in app_names:
            s.menu_paths.delete_menu_path(
                uuid=s.menu_paths.get_menu_path(name=app_name)["id"]
            )

        delete_dashboard_if_exists(name)

        app_ids = []
        s.boards.create_board(name=name)
        for app_name in app_names:
            app_id = s.menu_paths.get_menu_path(name=app_name)["id"]
            app_ids.append(app_id)

        s.boards.group_menu_paths(menu_path_names=app_names, name=name)
        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)

        for app_id, app_name in zip(app_ids, app_names):
            assert app_id in dashboard_app_ids
            s.boards.remove_menu_path_from_board(menu_path_name=app_name, name=name)
            dashboard_app_ids.remove(app_id)

        assert len(dashboard_app_ids) == 0

        s.boards.group_menu_paths(menu_path_ids=app_ids, name=name)
        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)

        for app_id, app_name in zip(app_ids, app_names):
            assert app_id in dashboard_app_ids
            s.boards.remove_menu_path_from_board(menu_path_name=app_name, name=name)
            dashboard_app_ids.remove(app_id)

        assert len(dashboard_app_ids) == 0

        all_app_ids = s.workspaces.get_workspace_menu_path_ids(uuid=business_id)

        s.boards.group_menu_paths(menu_path_names="all", name=name)
        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)

        for app_id in all_app_ids:
            assert app_id in dashboard_app_ids
            s.boards.remove_menu_path_from_board(menu_path_id=app_id, name=name)
            dashboard_app_ids.remove(app_id)

        assert len(dashboard_app_ids) == 0

        s.boards.delete_board(name=name)

        assert not s.boards.get_board(name=name)

        for app_id in app_ids:
            s.menu_paths.delete_menu_path(uuid=app_id)

    def test_dashboard_theme(self):
        name = "Testing dashboard theme"

        delete_dashboard_if_exists(name)

        s.workspaces.update_workspace(
            uuid=business_id, theme={"palette": {"background": {"default": "#23232F"}}}
        )

        s.boards.create_board(name=name)

        board = s.boards.get_board(name=name)

        assert board["theme"] == {"palette": {"background": {"default": "#23232F"}}}

        s.boards.update_board(uuid=board["id"], theme={})
        board = s.boards.get_board(name=name)

        assert board["theme"] == {}

        s.boards.delete_board(name=name)

        assert not s.boards.get_board(name=name)

        s.workspaces.update_workspace(uuid=business_id, theme={})



from os import getenv
from shimoku import Client

# from mock_classes import MockClient


def initiate_shimoku():
    api_key: str = getenv("API_TOKEN")
    universe_id: str = getenv("UNIVERSE_ID")
    environment: str = getenv("ENVIRONMENT")
    verbose: str = getenv("VERBOSITY")
    mock: bool = getenv("MOCK") == "TRUE"
    async_execution: bool = getenv("ASYNC_EXECUTION") == "TRUE"
    playground: bool = getenv("PLAYGROUND") == "TRUE"
    local_port: int = int(getenv("LOCAL_PORT"))

    if playground:
        s = Client(
            environment=environment,
            verbosity=verbose,
            async_execution=async_execution,
            local_port=local_port,
            retry_attempts=1,
        )
    # elif mock:
    # s = MockClient(
    #     verbosity=verbose,
    #     async_execution=async_execution,
    # )
    else:
        s = Client(
            access_token=api_key,
            universe_id=universe_id,
            environment=environment,
            verbosity=verbose,
            async_execution=async_execution,
            retry_attempts=1,
        )

    return s



from typing import Optional, Union, TYPE_CHECKING

from shimoku.api.base_resource import Resource

if TYPE_CHECKING:
    from shimoku.api.resources.business import Business
    from shimoku.api.resources.app import App
    from shimoku.api.resources.dashboard import Dashboard

import logging

logger = logging.getLogger(__name__)


class Event(Resource):
    """Event resource class"""

    _module_logger = logger
    resource_type = "event"
    plural = "events"

    def __init__(self,
        parent: "Business",
        uuid: Optional[str] = None,
    ):
        params = dict(
            type="NO_EVENT",
            resourceId=None,
            content={},
            authResourceType="BUSINESS_ADMIN",
            mainAuthResourceId=parent["id"],
        )

        super().__init__(
            parent=parent, uuid=uuid, params=params, params_to_serialize=["content"]
        )

    def set_auth_resource(self, auth_resource: Union["App", "Dashboard"]):
        self["authResourceType"] = auth_resource.resource_type.upper()
        self["mainAuthResourceId"] = auth_resource["id"]



""""""
from os import getenv
from typing import Optional
import unittest
import time
import json
import datetime as dt
from shimoku.exceptions import ShimokuFileError
from shimoku import Client

access_token: str = getenv("UNIVERSE_API_TOKEN")
universe_id: str = getenv("UNIVERSE_ID")
business_id: str = getenv("BUSINESS_ID")
verbosity: str = getenv("VERBOSITY")
environment: str = getenv("ENVIRONMENT")

last_execution_time: float = 0


def execute_ai_function(wf: str, model_name: Optional[str] = None) -> str:
    global last_execution_time
    tics = 0
    while last_execution_time is not None and time.time() - last_execution_time < 15:
        print("\rWaiting for min run interval" + ("." * (tics % 4)) + "      ", end="")
        tics += 1
        time.sleep(0.5)
    print()
    s.ai.create_input_files(
        input_files={
            "Great Content": (
                b"//////////////////////////// CONTENT OF THE INPUT FILE ////////////////////////////\n\n"
                b"Great Content\n\n///////////////////////////////////////////////////////////////////////////////////",
                {"Great metadata entry": "important file info"},
            )
        },
        force_overwrite=True,
    )
    last_execution_time = time.time()
    params = dict(
        ai_function=wf,
        input_data="Great Content",
        text_mandatory="Important text",
        text_optional="Less important text",
        add_timestamp=True,
    )
    if model_name:
        params["model_name"] = model_name
    return s.ai.generic_execute(**params)


s = Client(
    access_token=access_token,
    universe_id=universe_id,
    verbosity=verbosity,
    environment=environment,
    retry_attempts=1,
)
s.set_workspace(uuid=business_id)

AI_FUNCTION_TO_TEST = "test_sdk_workflow_create_outputs"
AI_FUNCTION_ID_TO_TEST = s.activity_templates.get_activity_template(
    name_version=(AI_FUNCTION_TO_TEST, "")
)["id"]
MENU_PATH = "test-ai"
MODEL_NAME = "test-model"

if MENU_PATH in {
    menu["name"] for menu in s.workspaces.get_workspace_menu_paths(business_id)
}:
    s.menu_paths.delete_all_menu_path_files(name=MENU_PATH, with_shimoku_generated=True)
    s.menu_paths.delete_all_menu_path_activities(
        name=MENU_PATH, with_linked_to_templates=True
    )
    s.menu_paths.delete_menu_path(name=MENU_PATH)

s.set_menu_path(name=MENU_PATH)


def test_private_methods():
    run_id_for_model = execute_ai_function(AI_FUNCTION_TO_TEST)
    s.ai.check_for_private_access(
        ai_function_id=AI_FUNCTION_ID_TO_TEST, run_id=run_id_for_model
    )
    ai_function_methods = s.ai.get_private_ai_function_methods()
    ai_function_methods.create_model(
        model_name=MODEL_NAME, model=b"", metadata={"other_field": "other_value"}
    )
    model, metadata = ai_function_methods.get_model(model_name=MODEL_NAME)
    assert model == b""
    assert metadata == {
        "creator_ai_function": AI_FUNCTION_TO_TEST,
        "creator_ai_function_id": AI_FUNCTION_ID_TO_TEST,
        "creator_ai_function_version": "",
        "model_name": MODEL_NAME,
        "other_field": "other_value",
        "run_id": run_id_for_model,
    }
    s.ai.delete_model(model_name="test-model")

    class ModelDoesntExist(unittest.TestCase):
        def test_model_doesnt_exist(self):
            with self.assertRaises(ShimokuFileError):
                ai_function_methods.get_model(model_name=MODEL_NAME)

    t = ModelDoesntExist()
    t.test_model_doesnt_exist()

    # For the following tests the model must exist
    ai_function_methods.create_model(
        model_name=MODEL_NAME, model=b"", metadata={"other_field": "other_value"}
    )


def test_create_input_files():
    input_files = s.ai.get_available_input_files()
    if "file1" in [input_file["file_name"] for input_file in input_files]:
        s.ai.delete_input_file(file_name="file1")
    if "file2" in [input_file["file_name"] for input_file in input_files]:
        s.ai.delete_input_file(file_name="file2")
    len_files = len(s.menu_paths.get_menu_path_files(name=MENU_PATH))
    len_input_files = len(s.ai.get_available_input_files())
    s.ai.create_input_files(
        {"file1": b"", "file2": (b"", {"metadata_key": "metadata_value"})}
    )
    assert len(s.menu_paths.get_menu_path_files(name=MENU_PATH)) == len_files
    input_files = s.ai.get_available_input_files()
    assert len(input_files) == len_input_files + 2
    assert "file1" in [input_file["file_name"] for input_file in input_files]
    assert "file2" in [input_file["file_name"] for input_file in input_files]
    assert ("metadata_key", "metadata_value") in [
        input_file_metadata.items()
        for input_file_metadata in input_files
        if input_file_metadata["file_name"] == "file2"
    ][0]
    s.ai.delete_input_file(file_name="file1")
    s.ai.delete_input_file(file_name="file2")
    assert (
        len(
            s.menu_paths.get_menu_path_files(
                name=MENU_PATH, with_shimoku_generated=True
            )
        )
        == len_files + 2
    )
    assert len(s.ai.get_available_input_files()) == len_input_files


def test_execute_ai_function():
    run_id = execute_ai_function(AI_FUNCTION_TO_TEST, MODEL_NAME)
    first_time = time.time()
    activity_ended = False
    last_log = ""
    s.disable_caching()
    while time.time() - first_time < 60:
        run = s.ai.get_last_executions_with_logs(AI_FUNCTION_TO_TEST)[0]
        if last_log != run["logs"][-1]["message"]:
            last_log = run["logs"][-1]["message"]
            print(last_log)
            if last_log == "The Activity has ended":
                activity_ended = True
                break
    if not activity_ended:
        raise Exception("AI function did not finish")
    output_dict = s.ai.get_output_file_objects(run_id=run_id)
    s.enable_caching()
    file_name, (file_obj, metadata) = list(output_dict.items())[0]

    decoded_file = file_obj.decode("utf-8")
    assert "Important text" in decoded_file
    assert "Less important text" in decoded_file
    assert "Great Content" in decoded_file
    assert dt.datetime.now().strftime("%Y-%m-%d") in file_name
    assert metadata["model_name"] == MODEL_NAME

    def assert_output_dict(_output_dict: dict):
        if "model_name" in _output_dict:
            assert _output_dict["model_name"] == MODEL_NAME
        if "ai_function_name" in _output_dict:
            assert _output_dict["ai_function_name"] == AI_FUNCTION_TO_TEST
            assert _output_dict["ai_function_version"] == ""

        assert _output_dict["input"]["args"] == {
            "add_timestamp": True,
            "text_mandatory": "Important text",
            "text_optional": "Less important text",
            "model_name": MODEL_NAME,
        }
        assert _output_dict["input"]["files"] == {
            "input_data": {
                "file_name": "Great Content",
                "Great metadata entry": "important file info",
            }
        }
        assert _output_dict["output_files"][file_name]["model_name"] == MODEL_NAME

    output_dict_by_ai_function = [
        o_dict
        for o_dict in s.ai.get_output_files_by_ai_function(AI_FUNCTION_TO_TEST)
        if o_dict["run_id"] == run_id
    ][0]
    assert_output_dict(output_dict_by_ai_function)
    output_dict_by_model = [
        o_dict
        for o_dict in s.ai.get_output_files_by_model(MODEL_NAME)
        if o_dict["run_id"] == run_id
    ][0]
    assert_output_dict(output_dict_by_model)


test_private_methods()
test_create_input_files()
test_execute_ai_function()

s.ai.show_available_ai_functions()
s.ai.show_ai_function_parameters(AI_FUNCTION_TO_TEST)
s.ai.show_available_models()
s.ai.show_available_input_files()
print(json.dumps(s.ai.get_output_files_by_ai_function(AI_FUNCTION_TO_TEST), indent=4))
s.ai.show_last_execution_logs_by_ai_function(AI_FUNCTION_TO_TEST)
s.ai.show_last_execution_logs_by_model(MODEL_NAME)



import datetime as dt

table_data = [
    {
        "date": dt.date(2021, 1, 1),
        "x": 5,
        "y": 5,
        "filtA": "A",
        "filtB": "Z",
        "name": "Ana",
        "web": "https://www.google.com",
        "webName": "Google",
        "web2": "https://www.youtube.com",
        "webName2": "Youtube",
        "modal": "Modal 1",
    },
    {
        "date": dt.date(2021, 1, 2),
        "x": 6,
        "y": 5,
        "filtA": "B",
        "filtB": "Z",
        "name": "Laura",
        "web": "https://www.facebook.com",
        "webName": "Facebook",
        "web2": "https://www.twitter.com",
        "webName2": "Twitter",
        "modal": "Modal 2",
    },
    {
        "date": dt.date(2021, 1, 3),
        "x": 4,
        "y": 5,
        "filtA": "A",
        "filtB": "W",
        "name": "Audrey",
        "web": "https://www.twitter.com",
        "webName": "Twitter",
        "web2": "https://www.instagram.com",
        "webName2": "Instagram",
        "modal": "Modal 3",
    },
    {
        "date": dt.date(2021, 1, 4),
        "x": 7,
        "y": 5,
        "filtA": "B",
        "filtB": "W",
        "name": "Jose",
        "web": "https://www.instagram.com",
        "webName": "Instagram",
        "web2": "https://www.linkedin.com",
        "webName2": "Linkedin",
        "modal": "Modal 1",
    },
    {
        "date": dt.date(2021, 1, 5),
        "x": 3,
        "y": 5,
        "filtA": "A",
        "filtB": "Z",
        "name": "Jorge",
        "web": "https://www.linkedin.com",
        "webName": "Linkedin",
        "web2": "https://www.google.com",
        "webName2": "Google",
        "modal": "Modal 2",
    },
]



from typing import Optional, Union, TYPE_CHECKING

from copy import deepcopy

import pandas as pd

from shimoku.plt.utils import ShimokuPalette
from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


async def line_chart(
    self: "PlotLayer",
    data: Union[str, pd.DataFrame, list[dict]],
    order: int,
    x: str,
    y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a line chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options["type"] = "line"
    common_options["yAxis"][0]["scale"] = True
    common_options["xAxis"][0]["boundaryGap"] = False
    if variant == "minimal":
        series_options["itemStyle"] = {"opacity": 0}
        series_options["lineStyle"] = {"width": 3}

    await self._create_trend_chart(
        option_modifications=option_modifications,
        echart_options=common_options,
        series_options=series_options,
        variant=variant,
        axes=x,
        values=y,
        data_mapping_to_tuples=await self._choose_data(order, data),
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
    )


async def area_chart(
    self: "PlotLayer",
    data: Union[str, pd.DataFrame, list[dict]],
    order: int,
    x: str,
    y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create an area chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options.update(
        {
            "type": "line",
            "areaStyle": {"opacity": 0.5},
        }
    )
    if variant == "minimal":
        series_options["itemStyle"] = {"opacity": 0}
        series_options["areaStyle"]["opacity"] = 0.2
        series_options["lineStyle"] = {"width": 3}

    common_options["xAxis"][0]["boundaryGap"] = False
    await self._create_trend_chart(
        option_modifications=option_modifications,
        echart_options=common_options,
        series_options=series_options,
        axes=x,
        values=y,
        data_mapping_to_tuples=await self._choose_data(order, data),
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
        variant=variant,
    )


async def stacked_area_chart(
    self: "PlotLayer",
    data: Union[str, pd.DataFrame, list[dict]],
    order: int,
    x: str,
    y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a stacked area chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options.update(
        {
            "type": "line",
            "areaStyle": {},
            "stack": "stack",
            "itemStyle": {"borderRadius": [0, 0, 0, 0]},
        }
    )
    common_options["xAxis"][0]["boundaryGap"] = False
    data_mapping_to_tuples = await self._choose_data(order, data)
    len_y = len(y) if y else len(data_mapping_to_tuples.keys()) - 1
    series_options = [series_options] * (len_y - 1)
    series_options.append(deepcopy(series_options[-1]))
    series_options[-1]["itemStyle"]["borderRadius"] = [9, 9, 0, 0]
    await self._create_trend_chart(
        option_modifications=option_modifications,
        echart_options=common_options,
        series_options=series_options,
        axes=x,
        values=y,
        data_mapping_to_tuples=data_mapping_to_tuples,
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
        variant=variant,
    )


async def predictive_line_chart(
    self: "PlotLayer",
    data: Union[str, pd.DataFrame, list[dict]],
    order: int,
    x: str,
    y: Union[list[str], str, None] = None,
    min_value_mark: Optional[Union[str, int]] = None,
    max_value_mark: Optional[Union[str, int]] = None,
    color_mark: str = "rgba(255, 173, 177, 0.4)",
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a predictive line chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options["type"] = "line"
    common_options["yAxis"][0]["scale"] = True
    common_options["xAxis"][0]["boundaryGap"] = False
    series_options["markArea"] = {
        "itemStyle": {"color": color_mark},
        "data": [
            [
                {"name": "Prediction", "xAxis": min_value_mark},
                {"xAxis": max_value_mark},
            ]
        ],
    }
    await self._create_trend_chart(
        echart_options=common_options,
        series_options=series_options,
        option_modifications=option_modifications,
        axes=x,
        values=y,
        data_mapping_to_tuples=await self._choose_data(order, data),
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
        variant=variant,
    )


async def marked_line_chart(
    self: "PlotLayer",
    data: Union[str, pd.DataFrame, list[dict]],
    order: int,
    x: str,
    marks: list,
    y: Union[list[str], str, None] = None,
    color_mark: str = "rgba(255, 173, 177, 0.4)",
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a predictive line chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options["type"] = "line"
    common_options["yAxis"][0]["scale"] = True
    common_options["xAxis"][0]["boundaryGap"] = False
    series_options["markArea"] = {
        "itemStyle": {"color": color_mark},
        "data": [
            [
                {"name": name, "xAxis": min_value_mark},
                {"xAxis": max_value_mark},
            ]
            for name, min_value_mark, max_value_mark in marks
        ],
    }
    await self._create_trend_chart(
        echart_options=common_options,
        series_options=series_options,
        option_modifications=option_modifications,
        axes=x,
        values=y,
        data_mapping_to_tuples=await self._choose_data(order, data),
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
        variant=variant,
    )


def calculate_segments(df: pd.DataFrame, threshold: float, r: int, g: int, b: int):
    color_range = df["y"].max() - threshold
    df_as_dict = df.to_dict("records")
    segment = [None, None, None]
    segments = []
    for i in range(len(df_as_dict)):
        if df_as_dict[i]["y"] >= threshold and segment[0] is None:
            segment[0] = i - 1 if i > 0 else 0
        elif (df_as_dict[i]["y"] < threshold or i == len(df_as_dict) - 1) and segment[
            0
        ] is not None:
            if i < len(df_as_dict) - 1 and df_as_dict[i + 1]["y"] >= threshold:
                continue
            segment[1] = i
            max_from_segment = max(
                [val["y"] for val in df_as_dict[segment[0] : segment[1]]]
            )
            color_alpha = 0.3 + (
                0.7 * (max_from_segment - threshold) / color_range
                if color_range > 0
                else 0
            )
            segment[2] = f"rgba({r}, {g}, {b}, {color_alpha})"
            segments.append(tuple(segment))
            segment = [None, None, None]

    return segments


async def segmented_area_chart(
    self: "PlotLayer",
    data: Union[str, pd.DataFrame, list[dict]],
    order: int,
    x: str,
    y: str,
    segments: Optional[list] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    default_color: tuple[int, int, int] = (255, 0, 0),
    title: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
    top_area: bool = False,
    threshold: Optional[float] = None,
    labels: Optional[list[str]] = None,
):
    """Create a segmented area chart"""
    if not isinstance(y, str):
        raise TypeError("y must be string")
    if threshold is not None and segments is not None:
        raise ValueError("You cannot specify both threshold and segments")
    if threshold is None and segments is None:
        raise ValueError("You must specify either threshold or segments")
    data_mappings_to_tuples = await self._choose_data(order, data)
    if isinstance(data, str):
        data = self._shared_data[data]
    df = pd.DataFrame(data)
    if threshold is not None:
        segments = calculate_segments(df, threshold, *default_color)

    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    common_options["xAxis"][0]["boundaryGap"] = False
    common_options["legend"] = {"show": False}
    del common_options["toolbox"]["feature"]["magicType"]
    del common_options["toolbox"]["bottom"]
    common_options["toolbox"]["top"] = "0px"
    common_options["dataZoom"] = (
        [{"type": "inside"}, {"type": "slider"}] if variant is None else []
    )
    pieces = []
    for segment in segments:
        if len(segment) == 2:
            r, g, b = default_color
            pieces.append(
                {
                    "gte": segment[0],
                    "lte": segment[1],
                    "color": f"rgba({r}, {g}, {b}, 0.75)",
                }
            )
        elif len(segment) == 3:
            color = segment[2]
            if not isinstance(color, str):
                if not isinstance(color, tuple) or len(color) != 3:
                    log_error(
                        logger,
                        f"Color must be a string or a tuple of 3 elements, got {color}",
                        ValueError,
                    )
                color = f"rgba({color[0]}, {color[1]}, {color[2]}, 0.75)"
            pieces.append({"gte": segment[0], "lte": segment[1], "color": color})
        else:
            raise ValueError("Segment must be a tuple of 2 or 3 elements")
    common_options["visualMap"] = {
        "type": "piecewise",
        "show": False,
        "dimension": 0,
        "seriesIndex": 0,
        "pieces": pieces,
    }
    series_options.update(
        {
            "type": "line",
            "emphasis": {"disabled": True},
            "areaStyle": {"origin": "end" if top_area else "start"},
            "itemStyle": {"opacity": 0},
        }
    )

    series_options = [series_options, deepcopy(series_options)]
    del series_options[1]["areaStyle"]

    if labels is None:
        labels = []
    if len(labels) < len(segments):
        labels += [""] * (len(segments) - len(labels))
    if len(labels) > len(segments):
        labels = labels[: len(segments)]

    series_options[1]["markArea"] = {
        "itemStyle": {"color": "rgba(0, 0, 0, 0)"},
        "emphasis": {"disabled": True},
        "label": {
            "show": variant != "minimal",
            "emphasis": {"disabled": True},
            "borderColor": f"rgba({default_color[0]}, {default_color[1]}, {default_color[2]}, 1)",
            "borderWidth": 1,
            "borderRadius": 5,
            "padding": [5, 5, 5, 5],
            "backgroundColor": "rgba(0,0,0,0)",
        },
        "data": [
            [
                {"name": label, "xAxis": marks_tuple[0]},
                {"xAxis": marks_tuple[1]},
            ]
            for label, marks_tuple in zip(labels, segments)
            if label
        ],
    }

    await self._create_trend_chart(
        echart_options=common_options,
        series_options=series_options,
        option_modifications=option_modifications,
        axes=x,
        values=[y, y],
        data_mapping_to_tuples=data_mappings_to_tuples,
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
        variant=variant,
    )


async def line_with_confidence_area_chart(
    self: "PlotLayer",
    data: Union[str, pd.DataFrame, list[dict]],
    order: int,
    x: str,
    lower: str,
    y: str,
    upper: str,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    percentages: bool = False,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    common_options = get_common_echart_options()
    del common_options["toolbox"]["feature"]["magicType"]
    common_options.update(
        {
            "tooltip": {
                "trigger": "axis",
                "axisPointer": {
                    "type": "cross",
                    "animation": False,
                    "label": {
                        "backgroundColor": "#ccc",
                        "borderColor": "#aaa",
                        "borderWidth": 1,
                        "shadowBlur": 0,
                        "shadowOffsetX": 0,
                        "shadowOffsetY": 0,
                        "color": "#222",
                    },
                },
            },
            "xAxis": [
                {
                    "type": "category",
                    "fontFamily": "Rubik",
                    "nameLocation": "middle",
                    "boundaryGap": True,
                    "nameGap": 35,
                    "data": "#set_data#",
                }
            ],
            "yAxis": [
                {
                    "type": "value",
                    "fontFamily": "Rubik",
                    "splitNumber": 3,
                    "boundaryGap": True,
                    "nameLocation": "middle",
                    "nameGap": 60,
                    "axisLabel": {
                        "formatter": "{value}%" if percentages else "{value}"
                    },
                    "axisPointer": {
                        "label": {"formatter": "{value}%" if percentages else "{value}"}
                    },
                },
            ],
        }
    )
    series_options = [
        {
            "name": lower,
            "data": "#set_data#",
            "type": "line",
            "lineStyle": {"color": "#000", "opacity": 0.1},
            "areaStyle": {"color": "#000", "opacity": 0.1, "origin": "end"},
            "symbol": "none",
        },
        {
            "name": upper,
            "data": "#set_data#",
            "silent": True,
            "type": "line",
            "lineStyle": {"color": "#000", "opacity": 0.1},
            "areaStyle": {"color": "#000", "opacity": 0.1, "origin": "start"},
            "symbol": "none",
        },
        {
            "name": y,
            "data": "#set_data#",
            "type": "line",
            "itemStyle": {"color": ShimokuPalette.CHART_C1.value},
            "emphasis": {"lineStyle": {"color": ShimokuPalette.CHART_C1.value}},
            "showSymbol": False,
        },
    ]

    await self._create_trend_chart(
        data_mapping_to_tuples=await self._choose_data(order, data),
        order=order,
        axes=x,
        values=[lower, upper, y],
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        title=title,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        echart_options=common_options,
        series_options=series_options,
        option_modifications=option_modifications,
        variant=variant,
    )


async def segmented_line_chart(
    self: "PlotLayer",
    order: int,
    x: str,
    y: Union[str, list[str]],
    data: Optional[Union[str, pd.DataFrame, list[dict]]],
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    marking_lines: Optional[list[int]] = None,
    range_colors: Optional[list[str]] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a chart with bars and lines."""
    data_mappings_to_tuples = await self._choose_data(order, data)

    if isinstance(y, str):
        y = [y]

    if not marking_lines:
        marking_lines = [0]

    common_options = get_common_echart_options()
    common_options["xAxis"][0]["boundaryGap"] = False
    # Calculate the width of the visual map with the label of the marking line
    # ( label  legend_ini - legend_end spaces for the legend colors)
    common_options["grid"]["right"] = (
        max(
            [
                len(
                    f" {marking_lines[i + 1]}   {marking_lines[i]} - {marking_lines[i + 1]}        "
                )
                for i in range(0, len(marking_lines) - 1)
            ]
        )
        * 6
    )
    common_options.update(
        {
            "visualMap": {
                "top": 50,
                "right": 10,
                "itemHeight": 14 if variant != "minimal" else 0,
                "showLabel": variant != "minimal",
                "outOfRange": {"color": "#999"},
                "pieces": [
                    {"min": marking_lines[i], "max": marking_lines[i + 1]}
                    for i in range(len(marking_lines) - 1)
                ]
                + [{"min": marking_lines[-1]}],
            },
        }
    )
    if range_colors:
        common_options["visualMap"]["color"] = range_colors[::-1]
    series_options = get_common_series_options()
    series_options.update(
        {
            "type": "line",
            "color": ShimokuPalette.CHART_C5.value,
            "markLine": {
                "data": [{"yAxis": val} for val in marking_lines],
                "label": {"show": variant != "minimal"},
                "itemStyle": {"color": ShimokuPalette.CHART_C5.value, "opacity": 0.5},
            },
        }
    )

    await self._create_trend_chart(
        data_mapping_to_tuples=data_mappings_to_tuples,
        axes=x,
        values=y,
        echart_options=common_options,
        series_options=series_options,
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        option_modifications=option_modifications,
        variant=variant,
    )



from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.utils import calculate_percentages_from_list, interpret_color
from shimoku.exceptions import BentoboxError

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


async def shimoku_gauge_chart(
    self: "PlotLayer",
    order: int,
    value: Union[int, float],
    name: str = "",
    color: Optional[Union[str, int]] = 1,
    title: Optional[str] = None,
    rows_size: Optional[int] = 1,
    cols_size: Optional[int] = 3,
    padding: Optional[str] = None,
    is_percentage: bool = False,
    option_modifications: Optional[dict] = None,
):
    """Create a pie chart"""

    color = interpret_color(color)

    value_font_size = 24 * max(min(rows_size, int(cols_size / 2)), 1)

    if self._bentobox_data:
        value_font_size = 35 * max(
            int(
                min(
                    rows_size / 10,
                    (cols_size - self._bentobox_data["bentoboxSizeColumns"]) / 4,
                )
            ),
            1,
        )

    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options["yAxis"]
    del common_options["xAxis"]
    del common_options["grid"]
    del common_options["toolbox"]
    common_options["tooltip"]["trigger"] = "item"
    common_options["grid"] = {
        "left": "5%",
        "right": "5%",
        "top": "5%",
        "bottom": "5%",
        "containLabel": True,
    }
    series_options.update(
        {
            "type": "gauge",
            "startAngle": 180,
            "endAngle": 0,
            "radius": "100%",
            "min": 0,
            "max": 100,
            "pointer": {
                "show": False,
            },
            "progress": {
                "show": True,
                "width": 30 if rows_size > 1 else 25,
                "overlap": False,
                "roundCap": False,
                "clip": False,
                "itemStyle": {
                    "borderWidth": 0,
                    "borderColor": color,
                    "color": color,
                },
            },
            "splitLine": {
                "show": False,
            },
            "axisLine": {"lineStyle": {"width": 30 if rows_size > 1 else 25}},
            "axisTick": {"show": False},
            "axisLabel": {
                "show": False,
            },
            "title": {
                "fontSize": 16,
                "fontFamily": "Rubik",
                "offsetCenter": ["0", "30"],
            },
            "detail": {
                "fontSize": value_font_size,
                "fontFamily": "Rubik",
                "font": "inherit",
                "color": "#202A36",
                "borderColor": "auto",
                "borderWidth": 0,
                "formatter": ("-" if value < 0 else "")
                + "{value}"
                + ("%" if is_percentage else ""),
                "valueAnimation": True,
                "offsetCenter": ["0", "-10"],
            },
        }
    )
    common_options["series"] = [series_options]

    await self._create_echart(
        fields=[{"name": "name", "value": "value"}],
        options=common_options,
        option_modifications=option_modifications,
        data_mapping_to_tuples=await self._choose_data(
            order, DataFrame([{"name": name, "value": abs(value)}])
        ),
        order=order,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
    )


def shimoku_gauges_group(
    self: "PlotLayer",
    gauges_data: Union[DataFrame, list[dict]],
    order: int,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = 12,
    gauges_padding: Optional[str] = "3, 1, 1, 1",
    gauges_rows_size: Optional[int] = 9,
    gauges_cols_size: Optional[int] = 4,
    calculate_percentages: Optional[bool] = False,
):
    if isinstance(gauges_data, DataFrame):
        gauges_data = gauges_data.to_dict(orient="records")

    if calculate_percentages:
        percentages = calculate_percentages_from_list(
            [gauge["value"] for gauge in gauges_data], 0
        )
        for i in range(len(percentages)):
            gauges_data[i]["value"] = percentages[i]

    if self._bentobox_data:
        log_error(
            logger,
            "The gauges group uses a bentobox, so it cannot be used inside another bentobox. Please"
            " pop out of the current bentobox before using this function.",
            BentoboxError,
        )

    self.set_bentobox(cols_size, rows_size)
    for gauge in gauges_data:
        order += 1
        self.shimoku_gauge(
            value=gauge["value"],
            order=order,
            name=gauge.get("name"),
            padding=gauge["padding"] if gauge.get("padding") else gauges_padding,
            color=gauge["color"] if gauge.get("color") else 1,
            rows_size=gauge["rows_size"]
            if gauge.get("rows_size")
            else gauges_rows_size,
            cols_size=gauge["cols_size"]
            if gauge.get("cols_size")
            else gauges_cols_size,
            is_percentage=gauge["is_percentage"]
            if gauge.get("is_percentage")
            else True,
        )

    self.pop_out_of_bentobox()
    return order + 1



from shimoku import Client


def stacked_bar_chart(shimoku_client: Client):
    shimoku_client.set_menu_path("test-free-echarts", "stacked-bar-chart")

    shimoku_client.plt.stacked_bar(
        data="stacked data",
        x="Segment",
        x_axis_name="Distribution and weight of the Drivers",
        order=0,
        show_values=["Price"],
    )



from shimoku import Client
from .data.default import data


def chart_and_indicators(shimoku_client: Client):
    shimoku_client.set_menu_path("test-bentobox", "chart-and-indicators")
    indicator_groups = [
        [
            {
                "description": "-950.55",
                "title": "Dow Jones",
                "value": "33,195.92",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
            {
                "description": "-84.68",
                "title": "Nasdaq",
                "value": "10,852.16",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
            {
                "description": "-950.55",
                "title": "Dow Jones",
                "value": "33,195.92",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
            {
                "description": "-84.68",
                "title": "Nasdaq",
                "value": "10,852.16",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
            {
                "description": "-950.55",
                "title": "Dow Jones",
                "value": "33,195.92",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
        ],
        [
            {
                "description": "-950.55",
                "title": "Dow Jones",
                "value": "33,195.92",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
            {
                "description": "-84.68",
                "title": "Nasdaq",
                "value": "10,852.16",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
        ],
        [
            {
                "description": "Return of investment",
                "title": "ROI",
                "value": "1.5M",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
            {
                "description": "% of times the algorithm has predicted the relative position of "
                "NY prices with respect to HK prices correctly",
                "title": "Accuracy",
                "value": "76.67%",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
        ],
        [
            {
                "description": "Return of investment",
                "title": "ROI",
                "value": "1.5M",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
            {
                "description": "% of times the algorithm has predicted the relative position of "
                "NY prices with respect to HK prices correctly",
                "title": "Accuracy",
                "value": "76.67%",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
            {
                "description": "% of times the algorithm has predicted the relative position of "
                "NY prices with respect to HK prices correctly",
                "title": "Accuracy",
                "value": "76.67%",
                "align": "left",
                "color": "success",
                "variant": "contained",
            },
        ],
    ]

    shimoku_client.plt.chart_and_indicators(
        order=0,
        chart_rows_size=3,
        cols_size=6,
        chart_function=shimoku_client.plt.line,
        chart_parameters=dict(
            data=data,
            x="date",
            y="x",
            title="Line Chart With Indicators",
        ),
        indicators_groups=indicator_groups,
        indicators_parameters={},
    )



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame
from dataclasses import dataclass
import shimoku


class PlotLayerHeader:
    """
    This class is a high level abstraction of the API, it is used to create components and data sets easily.
    """

    @dataclass
    class TableButtonColumnDefinition:
        """
        column_name: str
        label: str
        modals_column: Optional[str] = None
        activities_column: Optional[str] = None
        actions_column: Optional[str] = None
        """

        column_name: str
        label: str
        modals_column: Optional[str] = None
        activities_column: Optional[str] = None
        actions_column: Optional[str] = None

    def action_button(
        self,
        label: str,
        order: int,
        action_id: str,
        rows_size: Optional[int] = 1,
        cols_size: int = 2,
        align: Optional[str] = "stretch",
        padding: Optional[str] = None,
    ):
        """
        Create a button in the dashboard that executes an action.
        :param label: the label of the button
        :param order: the order of the button
        :param action_id: the id of the action
        :param rows_size: the size of the rows in the button
        :param cols_size: the size of the columns in the button
        :param align: the alignment of the button
        :param padding: the padding of the button
        """
        pass

    def activity_button(
        self,
        label: str,
        order: int,
        rows_size: Optional[int] = 1,
        cols_size: int = 2,
        align: Optional[str] = "stretch",
        padding: Optional[str] = None,
        activity_id: Optional[str] = None,
        activity_name: Optional[str] = None,
    ):
        """
        Create a button in the dashboard that executes an activity.
        :param label: the label of the button
        :param order: the order of the button
        :param rows_size: the size of the rows in the button
        :param cols_size: the size of the columns in the button
        :param align: the alignment of the button
        :param padding: the padding of the button
        :param activity_id: the id of the activity
        :param activity_name: the name of the activity
        """
        pass

    def annotated_chart(
        self,
        data: Union[list[DataFrame], list[list[dict]]],
        order: int,
        x: str,
        y: Union[str, list[str]],
        annotations: str = "annotation",
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        slider_config: Optional[dict] = None,
        slider_marks: Optional[list[tuple]] = None,
    ):
        """
        Create an annotated chart in the menu path.
        :param data: the data of the chart
        :param order: the order of the chart
        :param x: the x axis
        :param y: the y axis
        :param annotations: the annotations
        :param rows_size: the rows size of the chart
        :param cols_size: the columns size of the chart
        :param padding: the padding of the chart
        :param title: the title of the chart
        :param y_axis_name: the name of the y axis
        :param slider_config: the slider config
        :param slider_marks: the slider marks
        """
        pass

    def area(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create an area chart"""
        pass

    def bar(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a bar chart"""
        pass

    def button(
        self,
        label: str,
        order: int,
        rows_size: Optional[int] = 1,
        cols_size: int = 2,
        align: Optional[str] = "stretch",
        padding: Optional[str] = None,
        on_click_events: Union[list[dict], dict, None] = None,
    ):
        """Create a button in the dashboard."""
        pass

    def change_current_tab(
        self,
        tab: str,
    ):
        """
        Change the current tab.
        :param tab: the name of the tab
        """
        pass

    def change_path(
        self,
        path: str,
    ):
        """Change the current path"""
        pass

    def chart_and_indicators(
        self,
        order: int,
        chart_parameters: dict,
        indicators_groups: list,
        indicators_parameters: Optional[dict] = None,
        chart_rows_size: int = 3,
        cols_size: int = 12,
        chart_function: Optional[callable] = None,
    ) -> int:
        """None"""
        pass

    def chart_and_modal_button(
        self,
        order: int,
        chart_parameters: dict,
        button_modal: str,
        rows_size: int = 3,
        cols_size: int = 12,
        chart_function: Optional[callable] = None,
        button_label: str = "Read more",
        button_side_text: str = "Click on the button to read more about this topic.",
    ):
        """None"""
        pass

    def clear_context(
        self,
    ):
        """None"""
        pass

    def clear_events_for_components(
        self,
    ):
        """Clear the events for the components"""
        pass

    def clear_menu_path(
        self,
    ):
        """
        Clear the current path or a subpath
        """
        pass

    def delete_chart_by_order(
        self,
        order: int,
    ):
        """
        Delete a report by order and context
        """
        pass

    def delete_modal(
        self,
        name: str,
    ):
        """Delete a modal"""
        pass

    def delete_tabs_group(
        self,
        name: str,
    ):
        """Delete a tabs group"""
        pass

    def doughnut(
        self,
        order: int,
        names: str,
        values: str,
        data: Union[str, DataFrame, list[dict]],
        title: Optional[str] = None,
        rows_size: Optional[int] = 2,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a doughnut chart"""
        pass

    def filter(
        self,
        order: int,
        data: str,
        field: str,
        multi_select: bool = False,
        cols_size: int = 4,
        rows_size: int = 1,
        padding: Optional[str] = None,
    ):
        """
        Filter the data set.
        :param order: the order of the chart
        :param data: the data
        :param field: the field to filter
        :param multi_select: whether to allow multi select
        :param cols_size: the size of the columns
        :param rows_size: the size of the rows
        :param padding: the padding
        """
        pass

    def free_echarts(
        self,
        data: Union[str, DataFrame, list[dict], None] = None,
        options: Optional[dict] = None,
        raw_options: Optional[str] = None,
        order: Optional[int] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        fields: Optional[list] = None,
        data_is_not_df: bool = False,
    ):
        """
        Create an echarts chart with custom options.
        :param data: the data of the chart
        :param options: the options of the chart
        :param raw_options: the raw options of the chart
        :param order: the order of the chart in the dashboard
        :param title: the title of the chart
        :param rows_size: the rows size of the chart
        :param cols_size: the columns size of the chart
        :param padding: the padding of the chart
        :param fields: the fields of the chart
        :param data_is_not_df: whether the data is not a dataframe
        """
        pass

    def funnel(
        self,
        order: int,
        names: str,
        values: str,
        data: Union[str, DataFrame, list[dict]],
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a scatter chart"""
        pass

    def gauge_indicator(
        self,
        order: int,
        value: int,
        title: Optional[str] = "",
        description: Optional[str] = "",
        cols_size: Optional[int] = 6,
        rows_size: Optional[int] = 1,
        color: Union[str, int] = 1,
    ):
        """Create a gauge indicator"""
        pass

    def generate_input_form_groups(
        self,
        order: int,
        form_groups: dict,
        dynamic_sequential_show: Optional[bool] = False,
        auto_send: Optional[bool] = False,
        next_group_label: Optional[str] = "Next",
        rows_size: Optional[int] = 3,
        cols_size: int = 12,
        padding: Optional[str] = None,
        modal: Optional[str] = None,
        activity_id: Optional[str] = None,
        activity_name: Optional[str] = None,
        action_id: Optional[str] = None,
        on_submit_events: Optional[list[dict]] = None,
    ):
        """Easier way to create an input form.
        :param menu_path: the menu path of the input form
        :param order: the order of the input form
        :param form_groups: the form groups of the input form
        :param dynamic_sequential_show: whether to show the next group after submitting the current group
        :param auto_send: whether to automatically send the form after the last group
        :param next_group_label: the label of the next group button
        :param rows_size: the rows size of the input form
        :param cols_size: the columns size of the input form
        :param padding: the padding of the input form
        :param modal: the modal to open after submitting the form
        :param activity_id: the activity id to run after submitting the form
        :param activity_name: the activity name to run after submitting the form
        :param action_id: the action id to run after submitting the form
        :param on_submit_events: the events to run after submitting the form
        """
        pass

    def get_component_by_order(
        self,
        order: int,
    ) -> Optional[dict]:
        """Get the component by order.
        :param order: the order of the component
        """
        pass

    def get_input_forms(
        self,
    ) -> list:
        """Get all the input forms in the current menu_path."""
        pass

    def get_modal(
        self,
        modal_name: str,
    ) -> Optional[dict]:
        """Get the modal.
        :param modal_name: the name of the modal
        """
        pass

    def get_shared_data_names(
        self,
    ) -> list:
        """Get the names of the shared data"""
        pass

    def get_tabs_group(
        self,
        tabs_index: tuple,
    ) -> Optional[dict]:
        """Get the tabs group.
        :param tabs_index: the index of the tabs group
        """
        pass

    def heatmap(
        self,
        order: int,
        x: str,
        y: str,
        values: str,
        data: Union[list[dict], DataFrame, str],
        color_range: Optional[tuple[int, int]] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        calculate_color_range: Optional[bool] = False,
        continuous: Optional[bool] = False,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a heatmap chart"""
        pass

    def horizontal_bar(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a horizontal bar chart"""
        pass

    def html(
        self,
        html: str,
        order: int,
        cols_size: Optional[int] = None,
        rows_size: Optional[int] = None,
        padding: Optional[str] = None,
    ):
        """
        Create an html report in the dashboard.
        :param html: the html code
        :param order: the order of the html
        :cols_size: the columns that the html occupies
        :rows_size: the rows that the html occupies
        :padding: padding
        """
        pass

    def iframe(
        self,
        url: str,
        order: int,
        height: int = 640,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
    ):
        """Create an iframe report in the dashboard.
        :param url: the url of the iframe
        :param order: the order of the iframe
        :param height: the height of the iframe
        :param cols_size: the columns that the iframe occupies
        :param padding: padding
        """
        pass

    def indicator(
        self,
        data: Union[str, DataFrame, list[dict], dict],
        order: int,
        vertical: Union[bool, str] = False,
        color_by_value: bool = False,
        cols_size: int = 12,
        rows_size: int = 1,
        padding: Optional[str] = None,
        **kwargs,
    ):
        """
        Create an indicator report in the dashboard.
        :param data: the data of the indicator
        :param order: the order of the indicator
        :param vertical: whether the indicator is vertical
        :param color_by_value: whether to color the indicator by value
        :cols_size: the columns that the indicator occupies
        :rows_size: the rows that the indicator occupies
        :padding: padding
        :title: the title of the indicator
        """
        pass

    def indicators_with_header(
        self,
        order: int,
        indicators_groups: list,
        title: str,
        indicators_parameters: Optional[dict] = None,
        subtitle: str = "",
        background_color: str = "var(--color-primary)",
        text_color: str = "var(--background-paper)",
        cols_size: int = 12,
        icon_url: str = "https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/63e3615716d4435d29e0b82c_Acurracy.svg",
    ) -> int:
        """None"""
        pass

    def infographics_text_bubble(
        self,
        title: str,
        text: str,
        order: int,
        chart_parameters: dict,
        rows_size: int = 3,
        cols_size: int = 12,
        chart_function: Optional[callable] = None,
        background_url: Optional[str] = None,
        background_color: str = "var(--background-default)",
        bubble_location: str = "top",
        image_url: Optional[str] = None,
        image_size: int = 100,
    ):
        """None"""
        pass

    def input_form(
        self,
        options: dict,
        order: int,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        modal: Optional[str] = None,
        activity_id: Optional[str] = None,
        activity_name: Optional[str] = None,
        action_id: Optional[str] = None,
        on_submit_events: Optional[list[dict]] = None,
    ):
        """Creates an input form.
        :param options: the options for the input form
        :param order: the order of the input form
        :param padding: the padding of the input form
        :param rows_size: the rows size of the input form
        :param cols_size: the columns size of the input form
        :param modal: the modal to open after submitting the form
        :param activity_id: the activity id to run after submitting the form
        :param activity_name: the activity name to run after submitting the form
        :param action_id: the action id to run after submitting the form
        :param on_submit_events: the events to run after submitting the form
        """
        pass

    def line(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a line chart"""
        pass

    def line_and_bar_charts(
        self,
        order: int,
        x: str,
        data: Union[str, DataFrame, list[dict], None],
        bar_names: Optional[list[str]] = None,
        line_names: Optional[list[str]] = None,
        x_axis_name: Optional[str] = None,
        bar_axis_name: Optional[str] = None,
        bar_suffix: Optional[str] = None,
        line_axis_name: Optional[str] = None,
        line_suffix: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a chart with bars and lines."""
        pass

    def line_with_confidence_area(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        lower: str,
        y: str,
        upper: str,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        percentages: bool = False,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """None"""
        pass

    def line_with_summary(
        self,
        order: int,
        data: Union[str, DataFrame, list[dict]],
        title: str,
        x: str,
        y: Union[str, list[str], None] = None,
        description: str = "",
        value: Union[str, float, int] = "",
        rows_size: int = 2,
        cols_size: int = 2,
    ):
        """None"""
        pass

    def marked_line(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        marks: list,
        y: Union[list[str], str, None] = None,
        color_mark: str = "rgba(255, 173, 177, 0.4)",
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a predictive line chart"""
        pass

    def modal_button(
        self,
        label: str,
        order: int,
        modal: str,
        rows_size: Optional[int] = 1,
        cols_size: int = 2,
        align: Optional[str] = "stretch",
        padding: Optional[str] = None,
    ):
        """
        Create a button in the dashboard that opens a modal.
        :param label: the label of the button
        :param order: the order of the button
        :param modal: the name of the modal
        :param rows_size: the size of the rows in the button
        :param cols_size: the size of the columns in the button
        :param align: the alignment of the button
        :param padding: the padding of the button
        """
        pass

    def pie(
        self,
        order: int,
        names: str,
        values: str,
        data: Union[str, DataFrame, list[dict]],
        title: Optional[str] = None,
        rows_size: Optional[int] = 2,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a pie chart"""
        pass

    def pop_out_of_bentobox(
        self,
    ):
        """Stop using a bentobox"""
        pass

    def pop_out_of_modal(
        self,
    ):
        """Pop the current modal."""
        pass

    def pop_out_of_tabs_group(
        self,
    ):
        """
        Pop the current tabs index.
        """
        pass

    def predictive_line(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        min_value_mark: Union[str, int, None] = None,
        max_value_mark: Union[str, int, None] = None,
        color_mark: str = "rgba(255, 173, 177, 0.4)",
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a predictive line chart"""
        pass

    def radar(
        self,
        order: int,
        names: str,
        data: Union[str, DataFrame, list[dict]],
        values: Optional[list[str]] = None,
        max_field: Optional[str] = None,
        fill_area: bool = False,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a radar chart"""
        pass

    def raise_if_cant_change_path(
        self,
    ):
        """Raise an error if a tabs group or a modal is already open."""
        pass

    def rose(
        self,
        order: int,
        names: str,
        values: str,
        data: Union[str, DataFrame, list[dict]],
        title: Optional[str] = None,
        rows_size: Optional[int] = 2,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a rose chart"""
        pass

    def sankey(
        self,
        order: int,
        sources: str,
        targets: str,
        values: str,
        data: list,
        option_modifications: Optional[dict] = None,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
    ):
        """Create a sankey chart"""
        pass

    def scatter(
        self,
        order: int,
        point_fields: Union[list[tuple[str, str]], tuple[str, str]],
        data: Union[str, DataFrame, list[dict]],
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a scatter chart"""
        pass

    def scatter_with_effect(
        self,
        order: int,
        x: str,
        y: str,
        data: Union[list[str], DataFrame, list[dict], None],
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        effect_points: Optional[list] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a scatter chart"""
        pass

    def segmented_area(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: str,
        segments: Optional[list] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        default_color: tuple = (255, 0, 0),
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
        top_area: bool = False,
        threshold: Optional[float] = None,
        labels: Optional[list[str]] = None,
    ):
        """Create a segmented area chart"""
        pass

    def segmented_line(
        self,
        order: int,
        x: str,
        y: Union[str, list[str]],
        data: Union[str, DataFrame, list[dict], None],
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        marking_lines: Optional[list[int]] = None,
        range_colors: Optional[list[str]] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a chart with bars and lines."""
        pass

    def send_events_for_components(
        self,
        orders: list,
    ):
        """Send events for the components in each order
        :param orders: the orders of the components
        """
        pass

    def set_bentobox(
        self,
        cols_size: int,
        rows_size: int,
        order: Optional[int] = None,
    ):
        """Start using a bentobox, the id and the order will be set when the bentobox is used for the first time
        :param cols_size: the number of columns in the bentobox
        :param rows_size: the number of rows in the bentobox
        :order: the order of the bentobox in the dashboard
        """
        pass

    def set_modal(
        self,
        modal_name: str,
        width: Optional[int] = None,
        height: Optional[int] = None,
        open_by_default: Optional[bool] = None,
    ) -> None:
        """Set the current modal.
        :param modal_name: the name of the modal
        :param width: the width of the modal
        :param height: the height of the modal
        :param open_by_default: whether the modal is open by default
        """
        pass

    def set_shared_data(
        self,
        dfs: dict = None,
        custom_data: dict = None,
    ):
        """
        Set shared data for the s.plt to use.
        :param dfs: the data frames
        :param custom_data: the custom data
        """
        pass

    def set_tabs_index(
        self,
        tabs_index: tuple,
        order: Optional[int] = None,
        parent_tabs_index: Optional[tuple[str, str]] = None,
        padding: Optional[str] = None,
        cols_size: Optional[int] = None,
        rows_size: Optional[int] = None,
        just_labels: Optional[bool] = None,
        sticky: Optional[bool] = None,
    ):
        """Set the current tabs index.
        :param tabs_index: the index of the tabs group
        :param order: the order of the tabs group in the dashboard
        :param parent_tabs_index: the index of the parent tabs group
        :param cols_size: the size of the columns in the tabs group
        :param padding: the padding of the tabs group
        :param rows_size: the size of the rows in the tabs group
        :param just_labels: whether to show just the labels in the tabs group
        :param sticky: whether to make the tabs group sticky
        """
        pass

    def shimoku_gauge(
        self,
        order: int,
        value: Union[int, float],
        name: str = "",
        color: Union[str, int, None] = 1,
        title: Optional[str] = None,
        rows_size: Optional[int] = 1,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        is_percentage: bool = False,
        option_modifications: Optional[dict] = None,
    ):
        """Create a pie chart"""
        pass

    def shimoku_gauges_group(
        self,
        gauges_data: Union[DataFrame, list[dict]],
        order: int,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = 12,
        gauges_padding: Optional[str] = "3, 1, 1, 1",
        gauges_rows_size: Optional[int] = 9,
        gauges_cols_size: Optional[int] = 4,
        calculate_percentages: Optional[bool] = False,
    ):
        """None"""
        pass

    def single_indicator(
        self,
        data: dict,
        order: int,
        cols_size: Optional[int] = None,
        rows_size: Optional[int] = None,
        padding: Optional[str] = None,
    ):
        """
        Create an indicator report in the dashboard.
        :param data: the data of the indicator
        :param order: the order of the indicator
        :cols_size: the columns that the indicator occupies
        :rows_size: the rows that the indicator occupies
        :padding: padding
        """
        pass

    def speed_gauge(
        self,
        order: int,
        name: str,
        value: int,
        min_value: int,
        max_value: int,
        title: Optional[str] = None,
        rows_size: Optional[int] = 2,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a speed gauge chart"""
        pass

    def stacked_area(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a stacked area chart"""
        pass

    def stacked_bar(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a stacked bar chart"""
        pass

    def stacked_horizontal_bar(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a stacked horizontal bar chart"""
        pass

    def sunburst(
        self,
        order: int,
        data: Union[str, dict, list],
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a sunburst chart"""
        pass

    def table(
        self,
        order: int,
        data: Union[str, DataFrame, list[dict]],
        columns: Optional[list[str]] = None,
        columns_button: bool = True,
        filters: bool = True,
        export_to_csv: bool = True,
        search: bool = True,
        page_size: int = 10,
        page_size_options: Optional[list[int]] = None,
        buttons_column_definition: Optional[
            shimoku.plt.plt_layer.PlotLayer.TableButtonColumnDefinition
        ] = None,
        initial_sort_column: Optional[str] = None,
        sort_descending: bool = False,
        columns_options: Optional[dict] = None,
        categorical_columns: Optional[list[str]] = None,
        label_columns: Optional[dict] = None,
        web_link_columns: Optional[dict[str, str]] = None,
        open_links_in_new_tab: bool = False,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
    ):
        """Create a table report in the dashboard.
        :param order: the order of the table
        :param data: the data of the table
        :param columns: the columns of the table
        :param columns_button: whether to show the columns button
        :param filters: whether to show the filters button
        :param export_to_csv: whether to show the export to csv button
        :param search: whether to show the search bar
        :param page_size: the number of rows per page
        :param page_size_options: the options for the number of rows per page
        :param buttons_column_definition: the definition of the buttons column
        :param initial_sort_column: the initial sorting column
        :param sort_descending: whether to sort descending by the initial sorting column
        :param columns_options: the options for the columns
        :param categorical_columns: the categorical columns
        :param label_columns: the label columns
        :param web_link_column: the column to use as web link
        :param open_link_in_new_tab: whether to open the web link in a new tab
        :param title: the title of the table
        :param padding: the padding of the table
        :param rows_size: the rows size of the table
        :param cols_size: the columns size of the table
        """
        pass

    def top_bottom_area(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: Optional[int] = None,
        x: str = "x",
        top_names: Optional[list[str]] = None,
        bottom_names: Optional[list[str]] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        top_axis_name: Optional[str] = None,
        bottom_axis_name: Optional[str] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """None"""
        pass

    def top_bottom_line(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        top_names: Optional[list[str]] = None,
        bottom_names: Optional[list[str]] = None,
        rows_size: int = 4,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        top_axis_name: Optional[str] = None,
        bottom_axis_name: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """None"""
        pass

    def tree(
        self,
        order: int,
        data: Union[str, dict, list[dict]],
        radial: bool = False,
        vertical: bool = False,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a tree chart"""
        pass

    def treemap(
        self,
        order: int,
        data: Union[dict, list, None] = None,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        option_modifications: Optional[dict] = None,
    ):
        """Create a treemap chart"""
        pass

    def waterfall(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        positive: str,
        negative: str,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_balance: bool = False,
        variant: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """None"""
        pass

    def zero_centered_bar(
        self,
        data: Union[str, DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """Create a zero centered bar chart"""
        pass



import tqdm

from copy import copy
from typing import Optional, List, Dict

from shimoku import App
from shimoku.utils import create_function_name

from shimoku.code_gen.tree_generation import CodeGenTree
from shimoku.code_gen.file_generator import CodeGenFileHandler
from shimoku.code_gen.business_code_gen.apps_code_gen.data_sets_code_gen.code_gen_from_data_sets import (
    code_gen_from_shared_data_sets,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.code_gen_from_reports import (
    code_gen_from_reports_tree,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_tabs import (
    code_gen_tabs_functions,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_modal import (
    code_gen_modals_functions,
)


import logging
from shimoku.execution_logger import ClassWithLogging

logger = logging.getLogger(__name__)


class AppCodeGen(ClassWithLogging):
    """Class for generating code from a menu path."""

    _module_logger = logger

    def __init__(self, app: App, output_path: str, pbar: Optional[tqdm.tqdm] = None):
        self._app = app
        self.app_f_name = "menu_path_" + create_function_name(app["name"])
        self._output_path = f"{output_path}/" f"{self.app_f_name}"
        self._actual_bentobox: Optional[Dict] = None
        self._imports_code_lines = [
            "import os",
            "from shimoku import Client",
        ]
        self._file_generator: CodeGenFileHandler = CodeGenFileHandler(self._output_path)
        self._code_gen_tree: CodeGenTree = CodeGenTree(
            app, self._file_generator, pbar=pbar
        )

    @staticmethod
    def code_gen_report_params(report: Dict) -> List[str]:
        """Generate code for the parameters of a report.
        :param report: report to generate code from
        :return: list of code lines
        """
        report_params_to_get = {
            "order": "order",
            "title": "title",
            "sizeColumns": "cols_size",
            "sizeRows": "rows_size",
            "sizePadding": "padding",
        }
        return [
            f"    {report_params_to_get[k]}="
            + (f'"{(report[k])}",' if isinstance(report[k], str) else f"{report[k]},")
            for k in report
            if k in report_params_to_get
        ]

    async def generate_code(self):
        """Use the resources in the API to generate code_lines for the SDK. Create a file in
        the specified path with the generated code_lines.
        """
        await self._code_gen_tree.generate_tree()
        if self._code_gen_tree.needs_pandas:
            self._imports_code_lines.extend(["import pandas as pd"])

        code_lines: List[str] = []
        function_calls_code_lines: List[str] = []

        shared_data_sets_code_lines = await code_gen_from_shared_data_sets(self)
        scripts_imports = copy(self._imports_code_lines)
        for path in self._code_gen_tree.tree:
            function_code_lines = await code_gen_from_reports_tree(self, path)
            path_name = "sub_path_" + create_function_name(path) if path else "no_path"
            script_code_lines = [
                *scripts_imports,
                "",
                "",
                'data_folder_path = os.path.dirname(os.path.abspath(__file__)) + "/data"',
                *await code_gen_tabs_functions(self, path),
                *await code_gen_modals_functions(self, path),
                "",
                "",
                f"def {path_name}(shimoku_client: Client):",
                *["    " + line for line in function_code_lines],
                "",
            ]
            self._file_generator.generate_script_file(path_name, script_code_lines)
            self._imports_code_lines.append(f"from .{path_name} import {path_name}")
            function_calls_code_lines.extend(
                [
                    "",
                    f'shimoku_client.set_menu_path("{self._app["name"]}"'
                    + (f', "{path}")' if path is not None else ")"),
                    f"{path_name}(shimoku_client)",
                ]
            )

        main_code_lines = [
            f'shimoku_client.set_menu_path("{self._app["name"]}")',
            "shimoku_client.plt.clear_menu_path()",
            *shared_data_sets_code_lines,
            *function_calls_code_lines,
        ]

        code_lines.extend(
            [
                *self._imports_code_lines,
                "",
                "",
                'data_folder_path = os.path.dirname(os.path.abspath(__file__)) + "/data"',
                "",
                "",
                f"def {self.app_f_name}(shimoku_client: Client):",
                *["    " + line for line in main_code_lines],
                "",
            ]
        )

        self._file_generator.generate_script_file("app", code_lines)
        # Create an __init__.py file for the imports to work
        self._file_generator.generate_script_file("__init__", [""])



from typing import Optional

from shimoku.api.resources.app import App
from shimoku.api.resources.data_set import DataSet
from shimoku.api.resources.report import Report
from shimoku.api.resources.reports.modal import Modal
from shimoku.api.resources.reports.tabs_group import TabsGroup
from shimoku.api.utils import change_data_set_name_with_report

from shimoku.code_gen.file_generator import CodeGenFileHandler

import tqdm
import pandas as pd
import asyncio

import logging
from shimoku.execution_logger import ClassWithLogging

logger = logging.getLogger(__name__)


class CodeGenTree(ClassWithLogging):
    """Class for generating code generation tree."""

    _module_logger = logger

    def __init__(
        self,
        app: App,
        file_generator: CodeGenFileHandler,
        pbar: Optional[tqdm.tqdm] = None,
    ):
        self._app = app
        self._file_generator = file_generator
        self._pbar = pbar
        self.custom_data_sets_with_data = {}
        self.actual_bentobox: Optional[dict] = None
        self.all_tab_groups: dict[str, list[dict]] = {}
        self.tree: Optional[dict] = None
        self.shared_data_sets: list[str] = []
        self.needs_pandas: bool = False

    def _update_pbar(self, n: int = 1):
        if self._pbar is not None:
            self._pbar.update(n)

    async def _check_for_shared_data_sets(
        self,
        report: Report,
        seen_data_sets: set,
        individual_data_sets: dict[str, Report],
    ):
        """Check for shared data sets in a report.
        :param report: report to check
        :param seen_data_sets: set of data sets already seen
        :param individual_data_sets: list of data sets to append to
        """
        report_data_sets: list[
            Report.ReportDataSet
        ] = await report.get_report_data_sets()
        data_sets_from_rds = set([rds["dataSetId"] for rds in report_data_sets])

        for ds_id in data_sets_from_rds:
            if ds_id in self.shared_data_sets:
                continue
            if ds_id in seen_data_sets:
                self.shared_data_sets.append(ds_id)
                del individual_data_sets[ds_id]
            else:
                individual_data_sets[ds_id] = report
                seen_data_sets.add(ds_id)

    async def _check_for_create_data_set(
        self, data_set: DataSet, report: Optional[Report] = None
    ):
        """Create a file for a data set.
        :param data_set: data set to create file for
        :param report: report to create file for
        """
        if data_set["name"] is None:
            data_set["name"] = data_set["id"]

        data: list[dict] = [
            {
                k: v
                for k, v in dp.cascade_to_dict().items()
                if k not in ["id", "dataSetId"] and v is not None
            }
            for dp in await data_set.get_data_points()
        ]

        if len(data) == 0:
            return

        data_as_df = pd.DataFrame(data)

        if len(data) > 1 or "customField1" not in data[0]:
            output_name = (
                data_set["name"]
                if report is None
                else change_data_set_name_with_report(data_set, report)
            )
            reverse_columns = (
                {v: k for k, v in data_set["columns"].items()}
                if data_set["columns"]
                else None
            )
            self._file_generator.create_data_frame_file(
                output_name, data_as_df, reverse_columns
            )
            self.needs_pandas = True
        else:
            self.custom_data_sets_with_data[data_set["id"]] = data[0]["customField1"]

    async def _get_data_sets(self):
        """Create files for the data sets."""
        reports = await self._app.get_reports()

        # To store the data sets in the cache for the reports to have faster access to them
        await self._app.get_data_sets()

        individual_data_sets: dict[str, Report] = {}
        seen_data_sets = set()

        await asyncio.gather(
            *[
                self._check_for_shared_data_sets(
                    report, seen_data_sets, individual_data_sets
                )
                for report in reports
            ]
        )
        tasks = []
        for ds_id in self.shared_data_sets:
            tasks.append(
                self._check_for_create_data_set(await self._app.get_data_set(ds_id))
            )
        for ds_id, report in individual_data_sets.items():
            tasks.append(
                self._check_for_create_data_set(
                    await self._app.get_data_set(ds_id), report
                )
            )
        await asyncio.gather(*tasks)

    async def _tree_from_tabs_group(
        self,
        tree: list,
        tabs_group: TabsGroup,
        parent_tabs_index: tuple[str, str] = None,
    ):
        """Recursively build a tree of reports from a tabs group.
        :param tree: list to append to
        :param tabs_group: tabs group to build tree from
        :param parent_tabs_index: tuple of (tabs_group hash, tab name) of parent tabs group
        """
        path = tabs_group["path"] + "_" if tabs_group["path"] else ""
        name = tabs_group["properties"]["hash"][len(path) :]
        tabs_group_dict = {
            "tabs_group": tabs_group,
            "tabs": {},
            "order": tabs_group["order"],
            "parent_tabs_index": parent_tabs_index,
            "name": name,
        }

        path = tabs_group["path"]
        if path not in self.all_tab_groups:
            self.all_tab_groups[path] = []
        self.all_tab_groups[path].insert(0, tabs_group_dict)

        tree.append(tabs_group_dict)
        tabs = sorted(
            tabs_group["properties"]["tabs"].items(), key=lambda x: x[1]["order"]
        )
        for tab, tab_data in tabs:
            report_ids = tab_data["reportIds"]
            tab_dict = {"tab_groups": [], "other": []}
            tabs_group_dict["tabs"][tab] = tab_dict
            for child_id in report_ids:
                child_report = await self._app.get_report(child_id)
                if child_report["reportType"] == "TABS":
                    await self._tree_from_tabs_group(
                        tab_dict["tab_groups"],
                        child_report,
                        (tabs_group["properties"]["hash"], tab),
                    )
                else:
                    tab_dict["other"].append(child_report)
                self._update_pbar()

            tab_dict["tab_groups"] = sorted(
                tab_dict["tab_groups"], key=lambda x: x["tabs_group"]["order"]
            )
            tab_dict["other"] = sorted(tab_dict["other"], key=lambda x: x["order"])

    async def _tree_from_modal(self, tree: list, modal: Modal):
        """Recursively build a tree of reports from a modal.
        :param self: PlotApi instance
        :param tree: list to append to
        :param modal: modal to build tree from
        """
        path = modal["path"] + "_" if modal["path"] else ""
        name = modal["properties"]["hash"][len(path) :]
        modal_dict = {"modal": modal, "tab_groups": [], "other": [], "name": name}
        tree.append(modal_dict)
        for child_id in modal["properties"]["reportIds"]:
            child_report = await self._app.get_report(child_id)
            if child_report["reportType"] == "TABS":
                await self._tree_from_tabs_group(modal_dict["tab_groups"], child_report)
            else:
                modal_dict["other"].append(child_report)
            self._update_pbar()

        modal_dict["tab_groups"] = sorted(
            modal_dict["tab_groups"], key=lambda x: x["tabs_group"]["order"]
        )
        modal_dict["other"] = sorted(modal_dict["other"], key=lambda x: x["order"])

    async def generate_tree(self):
        """Generate a tree of reports from a list of reports.
        :param self: PlotApi instances
        """
        self.tree = {}
        reports = await self._app.get_reports()
        contained_reports = set()
        for report in reports:
            # TODO: make this not necessary
            if "hash" not in report["properties"]:
                report["properties"]["hash"] = "id_" + report["id"]
            if report["reportType"] == "MODAL":
                contained_reports.update(report["properties"]["reportIds"])
            elif report["reportType"] == "TABS":
                for tab in report["properties"]["tabs"]:
                    contained_reports.update(
                        report["properties"]["tabs"][tab]["reportIds"]
                    )
        reports = sorted(
            reports,
            key=lambda x: (
                x["pathOrder"] if x["pathOrder"] else 0,
                "0"
                if x["reportType"] == "MODAL"
                else "1"
                if x["reportType"] == "TABS"
                else "_" + (x["properties"]["hash"]),
            ),
        )
        for report in reports:
            if report["id"] in contained_reports:
                continue

            if report["path"] not in self.tree:
                self.tree[report["path"]] = {
                    "modals": [],
                    "tab_groups": [],
                    "other": [],
                }

            if report["reportType"] == "MODAL":
                await self._tree_from_modal(self.tree[report["path"]]["modals"], report)
            elif report["reportType"] == "TABS":
                await self._tree_from_tabs_group(
                    self.tree[report["path"]]["tab_groups"], report
                )
            else:
                self.tree[report["path"]]["other"].append(report)

            self._update_pbar()

        for path in self.tree:
            self.tree[path]["other"] = sorted(
                self.tree[path]["other"], key=lambda x: x["order"]
            )
            self.tree[path]["tab_groups"] = sorted(
                self.tree[path]["tab_groups"], key=lambda x: x["tabs_group"]["order"]
            )

        # Todo: Solve path ordering
        # self.tree = {k: v for k, v in sorted(reports.items(), key=lambda item: }
        await self._get_data_sets()



free_echarts_raw_options = """
    {title: {
        text: 'Stacked Area Chart'
      },
      tooltip: {
        trigger: 'axis',
        axisPointer: {
          type: 'cross',
          label: {
            backgroundColor: '#6a7985'
          }
        }
      },
      legend: {
        data: ['Email', 'Union Ads', 'Video Ads', 'Direct']
      },
      toolbox: {
        feature: {
          saveAsImage: {}
        }
      },
      grid: {
        left: '3%',
        right: '4%',
        bottom: '3%',
        containLabel: true
      },
      xAxis: [
        {
          type: 'category',
          boundaryGap: false,
          data: ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
        }
      ],
      yAxis: [
        {
          type: 'value'
        }
      ],
      series: [
        {
          name: 'Email',
          type: 'line',
          stack: 'Total',
          areaStyle: {},
          emphasis: {
            focus: 'series'
          },
          data: [120, 132, 101, 134, 90, 230, 210]
        },
        {
          name: 'Union Ads',
          type: 'line',
          stack: 'Total',
          areaStyle: {},
          emphasis: {
            focus: 'series'
          },
          data: [220, 182, 191, 234, 290, 330, 310]
        },
        {
          name: 'Video Ads',
          type: 'line',
          stack: 'Total',
          areaStyle: {},
          emphasis: {
            focus: 'series'
          },
          data: [150, 232, 201, 154, 190, 330, 410]
        },
        {
          name: 'Direct',
          type: 'line',
          stack: 'Total',
          areaStyle: {},
          emphasis: {
            focus: 'series'
          },
          data: [320, 332, 301, 334, 390, 330, 320]
        },
      ]
    }
"""



""""""
from os import getenv
import unittest

import pandas as pd

from shimoku.exceptions import CacheError
from shimoku import Client


api_key: str = getenv("API_TOKEN")
universe_id: str = getenv("UNIVERSE_ID")
business_id: str = getenv("BUSINESS_ID")
environment: str = getenv("ENVIRONMENT")
verbosity: str = getenv("VERBOSITY")

s = Client(
    access_token=api_key,
    universe_id=universe_id,
    environment=environment,
    verbosity="INFO",
    async_execution=True,
    retry_attempts=1,
)
s.set_workspace(uuid=business_id)
s.set_menu_path("File test")
s.menu_paths.delete_all_menu_path_files(name="File test", with_shimoku_generated=True)


def test_general_files():
    data = {
        "a": [1, 2, 3],
        "b": [1, 4, 9],
    }
    df = pd.DataFrame(data)
    file_object = df.to_csv(index=False).encode("utf-8")
    filename = "helloworld.csv"
    s.io.post_object(file_name=filename, obj=file_object)

    files = s.menu_paths.get_menu_path_files(name="File test")
    assert "helloworld.csv" in [file["name"] for file in files]

    class MyTestCase(unittest.TestCase):
        def test_cant_create_duplicate_files(self):
            with self.assertRaises(CacheError):
                s.io.post_object(file_name=filename, obj=file_object, overwrite=False)
                s.run()

    t = MyTestCase()
    t.test_cant_create_duplicate_files()

    s.io.delete_file(file_name=filename)
    files = s.menu_paths.get_menu_path_files(name="File test")
    assert "helloworld.csv" not in [file["name"] for file in files]


def test_tags_and_metadata():
    previous_len_files = len(s.menu_paths.get_menu_path_files(name="File test"))
    s.io.post_object(
        file_name="Shimoku generated file",
        obj=b"",
        tags=["shimoku_generated", "IO"],
        metadata={"business_id": business_id},
    )
    len_files = len(s.menu_paths.get_menu_path_files(name="File test"))
    assert len_files == previous_len_files
    len_files = len(
        s.menu_paths.get_menu_path_files(name="File test", with_shimoku_generated=True)
    )
    assert len_files == previous_len_files + 1
    file = s.io.get_object(file_name="Shimoku generated file")
    assert file == b""
    file_metadata = s.io.get_file_metadata(file_name="Shimoku generated file")
    assert file_metadata["metadata"]["business_id"] == business_id
    assert file_metadata["tags"] == ["shimoku_generated", "IO"]
    s.io.delete_file(file_name="Shimoku generated file", with_shimoku_generated=True)
    len_files = len(
        s.menu_paths.get_menu_path_files(name="File test", with_shimoku_generated=True)
    )
    assert len_files == previous_len_files


def test_get_object():
    file_name = "helloworld"
    object_data = b""
    s.io.post_object(file_name=file_name, obj=object_data)
    file = s.io.get_object(file_name=file_name)
    assert file == object_data


def test_post_dataframe():
    file_name: str = "df-test"
    df = pd.DataFrame({"a": [1, 2, 3], "b": [1, 4, 9]})
    s.io.post_dataframe(file_name, df=df)
    file = s.io.get_dataframe(file_name=file_name)
    assert file.to_dict() == df.to_dict()


def test_post_get_model():
    from sklearn import svm
    from sklearn import datasets

    clf = svm.SVC()
    X, y = datasets.load_iris(return_X_y=True)
    clf.fit(X, y)

    s.io.post_ai_model(model_name="model-test", model=clf)
    clf2 = s.io.get_ai_model(model_name="model-test")

    assert type(clf2) == type(clf)


def test_big_data():
    filename = "../../data/bulidata.csv"
    df = pd.read_csv(filename)
    df.reset_index(inplace=True)

    s.io.post_batched_dataframe(file_name="test-big-df", df=df)

    dataset: pd.DataFrame = s.io.get_batched_dataframe(file_name="test-big-df")
    assert len(dataset) == len(df)

    s.io.post_object(
        file_name="test-big-df.csv", obj=df.to_csv(index=False).encode("utf-8")
    )


def test_more_than_100_files():
    menu_path_mt100 = "more than 100 files test"
    s.set_menu_path(menu_path_mt100)
    s.menu_paths.delete_all_menu_path_files(name=menu_path_mt100)
    for i in range(201):
        s.io.post_object(
            file_name=f"file_{i}.txt", obj=b"Content of file " + str(i).encode("utf-8")
        )
    s.pop_out_of_menu_path()
    s.run()
    s.disable_caching()
    assert len(s.menu_paths.get_menu_path_files(name=menu_path_mt100)) == 201
    s.menu_paths.delete_all_menu_path_files(name=menu_path_mt100)
    assert len(s.menu_paths.get_menu_path_files(name=menu_path_mt100)) == 0
    s.menu_paths.delete_menu_path(name=menu_path_mt100)


test_general_files()
test_get_object()
test_post_dataframe()
test_tags_and_metadata()
test_post_get_model()
# test_big_data()
test_more_than_100_files()

s.run()



horizontal_bar_data = [
    {"name": "Financiado", "value": 1321},
    {"name": "Product name 2", "value": 1400},
    {"name": "Product name 3", "value": 1499},
    {"name": "Policy tenure", "value": 1550},
    {"name": "Family size", "value": 1663},
    {"name": "Acquisitive power", "value": 1995},
    {"name": "Geo", "value": 2712},
    {"name": "Modelo", "value": 3521},
    {"name": "Año", "value": 4643},
    {"name": "Offer", "value": 5426},
]



from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.cloud.cascade_get_resources import InitOptions, ResourceGetter
from typing import Optional
from shimoku.code_gen.main_code_gen import generate_code as gen_code
from shimoku.execution_logger import log_error
from shimoku.utils import create_function_name
from shimoku import Business
import tqdm
import tempfile
import shutil
import subprocess

import logging

logger = logging.getLogger(__name__)


def add_persistence_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the playground parser to a parser
    :param parser: Parser to add the playground parser to
    """

    params = {
        "name": "persist",
        "description": "Commands to persist the contents of the current database",
    }

    if parser:
        persistence_parser = parser.add_command(**params)
    else:
        persistence_parser = CLIParser(**params)

    module_functions = [generate_code, commit, pull]

    common_args = [
        CLIFuncParam(
            name="local-port",
            arg_type=int,
            arg_help="Local port to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="environment",
            arg_type=str,
            arg_help="Environment to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="access-token",
            arg_type=str,
            arg_help="Access token to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="universe-id",
            arg_type=str,
            arg_help="Universe ID to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="workspace-id",
            arg_type=str,
            arg_help="Workspace ID to use",
            mandatory=False,
        ),
    ]

    for func in module_functions:
        persistence_parser.decor_add_command(common_args=common_args)(func)

    return persistence_parser


async def generate_code(
    output_path: Optional[str],
    menu_paths: Optional[list[str]],
    hide_progress_bar: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    use_black_formatter: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """
    Generate code for a set workspace.
    :param menu_paths: List of menu paths to generate code from, leave empty to generate code for all menu paths.
    :param use_black_formatter: Use black formatter to format the generated code.
    :param output_path: Output path for the generated code.
    :param hide_progress_bar: Show progress bar while generating code.
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    business_object = await resource_getter.get_business()

    if not output_path:
        output_path = "generated_code"
    pbar = None
    if not hide_progress_bar:
        pbar = tqdm.tqdm(
            total=sum(
                [
                    len(await app.get_reports())
                    for app in await business_object.get_apps()
                ]
            )
            + 1
        )
    await gen_code(
        business_object=business_object,
        output_path=output_path,
        menu_paths=menu_paths,
        use_black_formatter=use_black_formatter,
        pbar=pbar,
    )
    if pbar:
        pbar.close()


async def commit(
    target_workspace: str = CLIFuncParam(prompt=True),
    target_access_token: str = CLIFuncParam(prompt=True),
    target_universe_id: str = CLIFuncParam(prompt=True),
    target_environment: str = CLIFuncParam(default="production", mandatory=False),
    hide_progress_bar: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """
    Commit contents to another workspace.
    :param target_workspace: Workspace id or name to commit to.
    :param target_access_token: Access token with access to the target workspace.
    :param target_universe_id: Universe id of the target workspace.
    :param target_environment: Environment of the target workspace.
    :param hide_progress_bar: Show progress bar while generating code.
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    business_object = await resource_getter.get_business()
    try:
        target_resource_getter = ResourceGetter(
            InitOptions(
                universe_id=target_universe_id,
                access_token=target_access_token,
                environment=target_environment,
                workspace=target_workspace,
            )
        )
        target_business_object = await target_resource_getter.get_business()
    except Exception as e:
        log_error(logger, f"Could not access target workspace: {e}", RuntimeError)
        raise e  # IDE complains if this is not here, because it thinks that log_error is not raising an exception

    temp_dir = tempfile.mkdtemp()
    pbar = None
    if not hide_progress_bar:
        pbar = tqdm.tqdm(
            total=sum(
                [
                    len(await app.get_reports())
                    for app in await business_object.get_apps()
                ]
            )
            + 1
        )
    await gen_code(
        business_object=business_object,
        business_id=target_business_object["id"],
        access_token=target_access_token,
        universe_id=target_universe_id,
        environment=target_environment,
        output_path=temp_dir,
        menu_paths=None,
        use_black_formatter=False,
        pbar=pbar,
    )
    if pbar:
        pbar.close()
    process = subprocess.run(
        [
            "python",
            f'{temp_dir}/execute_workspace_{create_function_name(business_object["name"])}.py',
        ],
        check=True,
    )
    if process.returncode != 0:
        log_error(logger, "Error while executing the generated code.", RuntimeError)
    shutil.rmtree(temp_dir)


async def pull(
    origin_workspace: str = CLIFuncParam(prompt=True),
    origin_access_token: str = CLIFuncParam(prompt=True),
    origin_universe_id: str = CLIFuncParam(prompt=True),
    origin_environment: str = CLIFuncParam(default="production", mandatory=False),
    hide_progress_bar: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """
    Pull contents from another workspace.
    :param origin_workspace: Workspace id or name to pull from.
    :param origin_access_token: Access token with access to the origin workspace.
    :param origin_universe_id: Universe id of the origin workspace.
    :param origin_environment: Environment of the origin workspace.
    :param hide_progress_bar: Show progress bar while retrieving the necessary data.
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    business_object: Business = await resource_getter.get_business()
    try:
        origin_resource_getter = ResourceGetter(
            InitOptions(
                universe_id=origin_universe_id,
                access_token=origin_access_token,
                environment=origin_environment,
                workspace=origin_workspace,
            )
        )
        origin_business_object = await origin_resource_getter.get_business()
    except Exception as e:
        log_error(logger, f"Could not access origin workspace: {e}", RuntimeError)
        raise e  # IDE complains if this is not here, because it thinks that log_error is not raising an exception

    temp_dir = tempfile.mkdtemp()
    pbar = None
    if not hide_progress_bar:
        pbar = tqdm.tqdm(
            total=sum(
                [
                    len(await app.get_reports())
                    for app in await origin_business_object.get_apps()
                ]
            )
            + 1
        )
    await gen_code(
        business_object=origin_business_object,
        business_id=business_object["id"],
        access_token=business_object.api_client.access_token,
        universe_id=business_object.parent["id"],
        environment=business_object.api_client.environment,
        output_path=temp_dir,
        menu_paths=None,
        use_black_formatter=False,
        pbar=pbar,
    )
    if pbar:
        pbar.close()
    process = subprocess.run(
        [
            "python",
            f'{temp_dir}/execute_workspace_{create_function_name(origin_business_object["name"])}.py',
        ],
        check=True,
    )
    if process.returncode != 0:
        log_error(logger, "Error while executing the generated code.", RuntimeError)
    shutil.rmtree(temp_dir)



from typing import Optional
from os import getenv
from dataclasses import dataclass
import asyncio
from shimoku.cli.utils import get_profile_config
from shimoku.exceptions import (
    WorkspaceError,
    BoardError,
    MenuPathError,
    FileError,
    ActivityTemplateError,
    ActivityError,
    RoleError,
    WorkspaceUserError,
)
from shimoku.execution_logger import log_error
from shimoku.api.client import ApiClient
from shimoku.api.resources.universe import Universe
from shimoku.api.resources.business import Business
from shimoku.api.resources.business_user import BusinessUser
from shimoku.api.resources.dashboard import Dashboard
from shimoku.api.resources.activity_template import ActivityTemplate
from shimoku.api.resources.app import App
from shimoku.api.resources.activity import Activity
from shimoku.api.resources.file import File
from shimoku.api.resources.data_set import DataSet
from shimoku.api.resources.report import Report
from shimoku.api.resources.action import Action

from shimoku.api.user_access_classes.universes_layer import UniversesLayer
from shimoku.api.user_access_classes.actions_layer import ActionsLayer
from shimoku.api.user_access_classes.businesses_layer import WorkspacesLayer
from shimoku.api.user_access_classes.dashboards_layer import BoardsLayer
from shimoku.api.user_access_classes.apps_layer import MenuPathsLayer
from shimoku.api.user_access_classes.files_layer import FilesLayer
from shimoku.api.user_access_classes.activity_templates_layer import (
    ActivityTemplatesLayer,
)
from shimoku.api.user_access_classes.activities_layer import ActivitiesLayer
from shimoku.api.user_access_classes.data_sets_layer import DataSetsLayer
from shimoku.api.user_access_classes.reports_layer import ComponentsLayer

from shimoku.ai.ai_layer import AILayer
import logging

logger = logging.getLogger(__name__)


@dataclass
class InitOptions:
    KEY_MAP = {
        "environment": "ENVIRONMENT",
        "access_token": "ACCESS_TOKEN",
        "universe_id": "UNIVERSE_ID",
        "workspace_id": "WORKSPACE_ID",
    }

    access_token: Optional[str] = getenv("SHIMOKU_ACCESS_TOKEN")
    environment: Optional[str] = getenv("SHIMOKU_ENVIRONMENT")
    universe_id: Optional[str] = getenv("SHIMOKU_UNIVERSE_ID")
    workspace_id: Optional[str] = getenv("SHIMOKU_WORKSPACE_ID")
    action: Optional[str] = None
    board: Optional[str] = None
    menu_path: Optional[str] = None
    role: Optional[str] = None
    activity: Optional[str] = None
    file: Optional[str] = None
    component_id: Optional[str] = None
    data_set: Optional[str] = None
    local_port: int = getenv("SHIMOKU_PLAYGROUND_PORT") or 8000

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            self.set_value(key, value)

    def set_value(self, name: str, value: any):
        if value:
            setattr(self, name, value)

    def __getattribute__(self, item):
        attr = object.__getattribute__(self, item)
        if attr is None and item in InitOptions.KEY_MAP:
            config = get_profile_config()
            attr = config.get(InitOptions.KEY_MAP[item])
        return attr


class ResourceGetter:
    def __init__(self, init_opts: InitOptions):
        self.init_opts = init_opts
        self.universe: Optional[Universe] = None

    @staticmethod
    def get_resource_by_name_or_uuid(
        resource_list: list,
        reference: str,
        name_field: str = "name",
    ):
        return next(
            (
                resource
                for resource in resource_list
                if resource[name_field] == reference or resource["id"] == reference
            ),
            None,
        )

    async def get_api_client(self) -> ApiClient:
        api_client = ApiClient(
            environment=self.init_opts.environment,
            playground=self.init_opts.universe_id == "local",
            config={"access_token": self.init_opts.access_token},
            server_port=self.init_opts.local_port,
            retry_attempts=1,
        )
        api_client.semaphore = asyncio.Semaphore(
            api_client.semaphore_limit
        )  # need async for semaphore
        return api_client

    async def get_universes_layer(self) -> UniversesLayer:
        api_client: ApiClient = await self.get_api_client()
        return UniversesLayer(api_client=api_client)

    async def get_universe(self) -> Universe:
        if self.universe:
            return self.universe
        api_client: ApiClient = await self.get_api_client()
        api_client.semaphore = asyncio.Semaphore(api_client.semaphore_limit)
        return Universe(api_client, self.init_opts.universe_id)

    async def get_actions_layer(self) -> ActionsLayer:
        universe: Universe = await self.get_universe()
        return ActionsLayer(universe=universe)

    async def get_action(self) -> Action:
        universe: Universe = await self.get_universe()
        action_ref: Optional[str] = self.init_opts.action
        if not action_ref:
            log_error(logger, "Action not specified", ActivityTemplateError)
        action: Optional[Action] = self.get_resource_by_name_or_uuid(
            await universe.get_actions(), action_ref
        )
        if not action:
            log_error(logger, f"Action {action_ref} not found", ActivityTemplateError)
        return action

    async def get_activity_templates_layer(self) -> ActivityTemplatesLayer:
        universe: Universe = await self.get_universe()
        return ActivityTemplatesLayer(universe=universe)

    async def get_activity_template(self) -> ActivityTemplate:
        universe: Universe = await self.get_universe()
        activity_template_ref: Optional[str] = self.init_opts.activity_template
        if not activity_template_ref:
            log_error(logger, "Activity template not specified", ActivityTemplateError)
        activity_template: Optional[
            ActivityTemplate
        ] = self.get_resource_by_name_or_uuid(
            await universe.get_activity_templates(), activity_template_ref
        )
        if not activity_template:
            log_error(
                logger,
                f"Activity template {activity_template_ref} not found",
                ActivityTemplateError,
            )
        return activity_template

    @staticmethod
    def get_activity_template_fields_to_show(show_input_settings: bool) -> list[str]:
        fields = ["id", "name", "version"]
        if show_input_settings:
            fields.append("input_settings")
        return fields

    async def get_businesses_layer(self) -> WorkspacesLayer:
        universe: Universe = await self.get_universe()
        return WorkspacesLayer(universe=universe)

    async def get_business(self) -> Business:
        universe: Universe = await self.get_universe()
        business_id = self.init_opts.workspace_id
        if not business_id:
            log_error(logger, "Workspace not specified", WorkspaceError)
        business: Optional[Business] = await universe.get_business(business_id)
        if not business:
            log_error(logger, f"Workspace {business_id} not found", WorkspaceError)
        return business

    @staticmethod
    def get_business_fields_to_show(show_theme: bool) -> list[str]:
        fields = ["id", "name"]
        if show_theme:
            fields.append("theme")
        return fields

    @staticmethod
    def get_invitation_fields_to_show() -> list[str]:
        return ["email"]

    @staticmethod
    def get_user_fields_to_show() -> list[str]:
        return ["email", "name"]

    async def get_dashboards_layer(self) -> BoardsLayer:
        business: Business = await self.get_business()
        return BoardsLayer(business=business)

    async def get_dashboard(self) -> Dashboard:
        business: Business = await self.get_business()
        dashboard_ref: Optional[str] = self.init_opts.board
        if not dashboard_ref:
            log_error(logger, "Board not specified", BoardError)
        dashboard: Optional[Dashboard] = self.get_resource_by_name_or_uuid(
            await business.get_dashboards(), dashboard_ref
        )
        if not dashboard:
            log_error(logger, f"Board {dashboard_ref} not found", BoardError)
        return dashboard

    @staticmethod
    def get_dashboard_fields_to_show(
        show_public_permission: bool, show_theme: bool
    ) -> list[str]:
        fields = ["id", "name", "order"]
        if show_public_permission:
            fields.append("publicPermission")
        if show_theme:
            fields.append("theme")
        return fields

    async def get_apps_layer(self) -> MenuPathsLayer:
        business: Business = await self.get_business()
        return MenuPathsLayer(business=business)

    async def get_app(self, app_ref: Optional[str] = None) -> App:
        business: Business = await self.get_business()
        app_ref: Optional[str] = app_ref or self.init_opts.menu_path
        if not app_ref:
            log_error(logger, "Menu path not specified", MenuPathError)
        app: Optional[App] = self.get_resource_by_name_or_uuid(
            await business.get_apps(), app_ref
        )
        if not app:
            log_error(logger, f"Menu path {app_ref} not found", MenuPathError)
        return app

    @staticmethod
    def get_app_fields_to_show() -> list[str]:
        return [
            "id",
            "name",
            "normalizedName",
            "order",
            "hideTitle",
            "hidePath",
            "showBreadcrumb",
            "showHistoryNavigation",
        ]

    async def get_reports_layer(self) -> ComponentsLayer:
        app: App = await self.get_app()
        return ComponentsLayer(app=app)

    async def get_report(self) -> Report:
        app: App = await self.get_app()
        return await app.get_report(uuid=self.init_opts.component_id)

    @staticmethod
    def get_report_fields_to_show(show_all_fields: bool = False) -> list[str]:
        fields = [
            "id",
            "title",
            "path",
            "pathOrder",
            "reportType",
            "order",
            "sizeColumns",
            "sizeRows",
            "sizePadding",
        ]
        if show_all_fields:
            fields.extend(
                ["bentobox", "properties", "dataFields", "chartData", "smartFilters"]
            )
        return fields

    async def get_data_sets_layer(self) -> DataSetsLayer:
        app: App = await self.get_app()
        return DataSetsLayer(app=app)

    async def get_data_set(self) -> DataSet:
        app: App = await self.get_app()
        data_set_ref: Optional[str] = self.init_opts.data_set
        if not data_set_ref:
            log_error(logger, "Data set not specified", ActivityError)
        data_set: Optional[DataSet] = self.get_resource_by_name_or_uuid(
            await app.get_data_sets(), data_set_ref
        )
        if not data_set:
            log_error(logger, f"Data set ({data_set_ref}) not found", ActivityError)
        return data_set

    @staticmethod
    def get_data_set_fields_to_show() -> list[str]:
        return ["id", "name"]

    @staticmethod
    def get_data_fields_to_show(show_custom_field: bool) -> list[str]:
        fields = ["orderField1", "description"]
        for i in range(1, 6):
            fields.append(f"dateField{i}")
        for i in range(1, 51):
            fields.append(f"stringField{i}")
            fields.append(f"intField{i}")
        if show_custom_field:
            fields.append("customField1")
        return fields

    async def get_activities_layer(self) -> ActivitiesLayer:
        app: App = await self.get_app()
        return ActivitiesLayer(app=app)

    async def get_activity(self) -> Activity:
        app: App = await self.get_app()
        activity_ref: Optional[str] = self.init_opts.activity
        if not activity_ref:
            log_error(logger, "Activity not specified", ActivityError)
        activity: Optional[Activity] = self.get_resource_by_name_or_uuid(
            await app.get_activities(), activity_ref
        )
        if not activity:
            log_error(logger, f"Activity {activity_ref} not found", ActivityError)
        return activity

    @staticmethod
    def get_activity_fields_to_show(show_template: bool) -> list[str]:
        fields = ["id", "name", "settings"]
        if show_template:
            fields.extend(["activityTemplateWithMode", "universeApiKeyId"])
        return fields

    async def get_files_layer(self) -> FilesLayer:
        app: App = await self.get_app()
        return FilesLayer(app=app)

    async def get_file(self) -> File:
        app: App = await self.get_app()
        file_ref: Optional[str] = self.init_opts.file
        if not file_ref:
            log_error(logger, "File not specified", FileError)
        file: Optional[File] = self.get_resource_by_name_or_uuid(
            await app.get_files(), file_ref
        )
        if not file:
            log_error(logger, f"File {file_ref} not found", FileError)
        return file

    @staticmethod
    def get_file_fields_to_show() -> list[str]:
        return ["id", "name", "contentType", "tags", "metadata"]

    async def get_ai_layer(self) -> AILayer:
        app = await self.get_app()
        return AILayer(
            access_token=self.init_opts.access_token,
            universe=app.parent.parent,
            app=app,
        )

    async def get_role_dicts(self) -> list[dict]:
        if self.init_opts.menu_path:
            apps_layer = await self.get_apps_layer()
            app = await self.get_app()
            return await apps_layer.get_roles(app["id"])
        if self.init_opts.board:
            dashboards_layer = await self.get_dashboards_layer()
            dashboard = await self.get_dashboard()
            return await dashboards_layer.get_roles(dashboard["id"])

        business = await self.get_business()
        business_layer = await self.get_businesses_layer()
        return await business_layer.get_roles(business["id"])

    async def get_role_dict(self) -> dict:
        role_dicts = await self.get_role_dicts()
        role = self.init_opts.role
        role_obj = self.get_resource_by_name_or_uuid(role_dicts, role, "role")
        if not role_obj:
            log_error(logger, f"Role {role} not found", RoleError)
        return role_obj



from typing import Optional, Union, TYPE_CHECKING

import pandas as pd

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

from shimoku.exceptions import DataError

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


async def heatmap_chart(
    self: "PlotLayer",
    order: int,
    x: str,
    y: str,
    values: str,
    data: Union[list[dict], pd.DataFrame, str],
    color_range: Optional[tuple[int, int]] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    padding: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    calculate_color_range: Optional[bool] = False,
    continuous: Optional[bool] = False,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a heatmap chart"""
    color_range = (0, 10) if color_range is None else color_range
    if calculate_color_range:
        aux_data = data
        if isinstance(aux_data, str):
            if aux_data not in self._shared_data:
                log_error(
                    logger,
                    f"Shared data set ({aux_data}) not found in the menu path ({str(self._app)})",
                    DataError,
                )
            aux_data = self._shared_data[aux_data]
        if isinstance(aux_data, pd.DataFrame):
            aux_data = aux_data.to_dict("records")
        color_range = (
            min([row[values] for row in aux_data]),
            max([row[values] for row in aux_data]),
        )
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options["legend"]
    del common_options["toolbox"]["feature"]["magicType"]
    common_options["xAxis"][0].update({"type": "category", "data": None})
    common_options["yAxis"][0].update({"type": "category", "nameGap": 64})
    common_options["visualMap"] = {
        "calculable": True,
        "orient": "vertical",
        "type": "continuous" if continuous else "piecewise",
        "left": "right",
        "top": "center",
        "show": variant != "minimal",
        "min": color_range[0],
        "max": color_range[1],
    }
    common_options["grid"].update(
        {"top": "5%", "right": 80, "bottom": 28, "left": "5%" if y_axis_name else "1%"}
    )

    series_options.update(
        {
            "type": "heatmap",
            "itemStyle": {"borderRadius": [0, 0, 0, 0]},
        }
    )
    await self._create_trend_chart(
        data_mapping_to_tuples=await self._choose_data(order, data),
        axes=[],
        values=[(x, y, values)],
        echart_options=common_options,
        series_options=series_options,
        order=order,
        title=title,
        padding=padding,
        rows_size=rows_size,
        cols_size=cols_size,
        show_values="all",
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        option_modifications=option_modifications,
        variant=variant,
    )



from uuid import uuid1
from typing import Optional, TYPE_CHECKING

from shimoku.api.base_resource import Resource
from shimoku.api.resources.business_user import BusinessUser, BusinessInvitation
from shimoku.api.resources.role import (
    Role,
    create_role,
    get_role,
    get_roles,
    delete_role,
)
from shimoku.api.resources.event import Event
from shimoku.api.resources.dashboard import Dashboard
from shimoku.api.resources.app import App

from shimoku.utils import EventType
from shimoku.exceptions import WorkspaceError, MenuPathError, BoardError, RoleError

if TYPE_CHECKING:
    from shimoku.api.resources.universe import Universe

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


class Business(Resource):
    _module_logger = logger
    resource_type = "business"
    alias_field = "name"
    plural = "businesses"

    def __init__(
        self,
        parent: "Universe",
        uuid: Optional[str] = None,
        alias: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params = {
            "name": alias if alias else "",
            "theme": {},
        }

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            children=[Dashboard, App, Role, Event, BusinessUser, BusinessInvitation],
            check_params_before_creation=["name"],
            params_to_serialize=["theme"],
            params=params,
        )

        self.currently_in_use = False

    async def delete(self):
        if self.currently_in_use:
            log_error(
                logger,
                f"Workspace {str(self)} is currently in use and cannot be deleted",
                WorkspaceError,
            )
        return await self._base_resource.delete()

    async def update(self):
        return await self._base_resource.update()

    # Event methods
    async def create_event(
        self, event_type: EventType, content: dict, resource_id: Optional[str] = None
    ):
        """Creates an event but do not store it in the cache
        :param event_type: The type of the event
        :param content: The content of the event
        :param resource_id: The id of the resource that triggered the event
        """
        event = Event(parent=self)
        event.set_params(type=event_type.value, content=content, resourceId=resource_id)
        logger.info(f"Event {event_type.value} created")
        await event

    # Dashboard methods
    async def create_dashboard(
        self,
        name: str,
        order: int,
        is_public: bool = False,
        is_disabled: bool = False,
        theme: Optional[dict] = None,
    ) -> Dashboard:
        dashboard_metadata = dict(
            order=order,
            isDisabled=is_disabled,
            theme=theme or self["theme"],
        )
        if is_public:
            dashboard_metadata["publicPermission"] = dict(
                isPublic=True, permission="READ", token=str(uuid1())
            )
        dashboard = await self._base_resource.create_child(
            Dashboard, alias=name, **dashboard_metadata
        )
        logger.info(f'Board {name} created with id {dashboard["id"]}')
        return dashboard

    async def update_dashboard(
        self, uuid: Optional[str] = None, name: Optional[str] = None, **params
    ):
        if params.get("new_name") is not None:
            params["name"] = params.pop("new_name")
            params["new_alias"] = True
        dashboard = await self.get_dashboard(uuid, name, create_if_not_exists=False)
        if not dashboard:
            logger.warning(f"Board {name} not found, cannot update it")
            return
        await self._base_resource.update_child(
            Dashboard, uuid=dashboard["id"], **params
        )

    async def get_dashboard(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        create_if_not_exists: bool = True,
    ) -> Optional[Dashboard]:
        dashboard = await self._base_resource.get_child(Dashboard, uuid, name)

        if dashboard:
            logger.info(
                f'Retrieved board {dashboard["name"]} with id {dashboard["id"]}'
            )
        elif create_if_not_exists:
            if not name:
                log_error(logger, "Name is required to create a board", BoardError)
            how_many_dashboards = len(await self.get_dashboards())
            dashboard = await self._base_resource.create_child(
                Dashboard, alias=name, order=how_many_dashboards + 1
            )
            logger.info(f'Created board {dashboard["name"]} with id {dashboard["id"]}')

        return dashboard

    async def get_dashboards(self) -> list[Dashboard]:
        return await self._base_resource.get_children(Dashboard)

    async def delete_dashboard(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        dashboard = await self.get_dashboard(uuid, name, create_if_not_exists=False)
        result = await self._base_resource.delete_child(Dashboard, uuid, name)
        return result

    # App methods
    async def get_app(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        create_if_not_exists: bool = True,
    ) -> Optional[App]:
        app = await self._base_resource.get_child(App, uuid, name)

        if app:
            logger.info(f'Retrieved menu path {app["name"]} with id {app["id"]}')
        elif create_if_not_exists:
            if not name:
                log_error(
                    logger, "Name is required to create a menu path", MenuPathError
                )
            how_many_apps = len(await self.get_apps())
            app = await self._base_resource.create_child(
                App, alias=name, order=how_many_apps + 1
            )

            logger.info(f'Created menu path {app["name"]} with id {app["id"]}')

        return app

    async def get_apps(self) -> list[App]:
        return await self._base_resource.get_children(App)

    async def delete_app(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        app = await self.get_app(uuid, name, create_if_not_exists=False)
        result = await self._base_resource.delete_child(App, uuid, name)
        if self.api_client.playground:
            dashboards = await self.get_dashboards()
            for dashboard in dashboards:
                if app["id"] in await dashboard.list_app_ids():
                    await dashboard.remove_app(app)
        return result

    async def update_app(
        self, uuid: Optional[str] = None, name: Optional[str] = None, **params
    ):
        if params.get("new_name") is not None:
            params["name"] = params.pop("new_name")
            params["new_alias"] = True
        app = await self.get_app(uuid, name, create_if_not_exists=False)
        if not app:
            logger.warning(f"App {name} not found, cannot update it")
            return
        await self._base_resource.update_child(App, uuid=uuid, alias=name, **params)

    # Account methods
    async def invite_user(
        self, email: str, roles: list[str], send_email: bool
    ) -> BusinessInvitation:
        invitation = await self._base_resource.create_child(
            BusinessInvitation, email=email, roles=roles, sendEmail=send_email
        )
        logger.info(f"Invitation sent to ({email})")
        return invitation

    async def get_pending_invitations(self):
        return await self._base_resource.get_children(BusinessInvitation)

    async def delete_pending_invitation(self, email: str):
        return await self._base_resource.delete_child(BusinessInvitation, alias=email)

    async def get_users(self):
        return await self._base_resource.get_children(BusinessUser)

    async def delete_user(self, email: str):
        return await self._base_resource.delete_child(BusinessUser, alias=email)

    # Role methods
    get_role = get_role
    get_roles = get_roles
    create_role = create_role
    delete_role = delete_role
